# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from typing import Union
from pathlib import Path
from importlib import import_module

from towhee.hub.repo_manager import RepoManager
from towhee.utils.log import engine_log
from towhee.utils.yaml_utils import load_yaml, dump_yaml


class OperatorManager(RepoManager):
    """
    The Repo Manager to manage the operator repos.

    Args:
        author (`str`):
            The author of the repo.
        repo (`str`):
            The name of the repo.
        root (`str`):
            The root url where the repo located.
    """
    def __init__(self, author: str, repo: str, root: str = 'https://hub.towhee.io'):
        super().__init__(author, repo, root)
        # 2 represents operators when creating a repo in Towhee's hub
        self._class = 2

    def create(self, password: str) -> None:
        """
        Create a repo under current account.

        Args:
            password (`str`):
                Current author's password.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        if self.exists():
            engine_log.info('%s/%s repo already exists.', self._author, self._repo)
        else:
            self.hub_utils.create(password, self._class)

    def create_with_token(self, token: str) -> None:
        """
        Create a repo under current account.

        Args:
            token (`str`):
                Current author's token.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        if self.exists():
            engine_log.info('%s/%s repo already exists.', self._author, self._repo)
        else:
            self.hub_utils.create_repo(token, self._class)

    def init_nnoperator(self, file_temp: Union[str, Path], file_dest: Union[str, Path], framework: str = 'pytorch') -> None:
        """
        Initialize the files under file_dest by moving and updating the text under file_temp.

        Args:
            file_temp (`Union[str, Path]`):
                The path to the template files.
            file_dest (`Union[str, Path]`):
                The path to the local repo to init.
            framework (`str, Path`):
                The framework for the model, defaults to 'pytorch'

        Raises:
            (`HTTPError`)
                Raise error in request.
            (`OSError`)
                Raise error in writing file.
        """
        repo_temp = self._temp['nnoperator']

        ori_str_list = [f'author/{repo_temp}', repo_temp, ''.join(x.title() for x in repo_temp.split('-')), 'pytorch']
        tar_str_list = [f'{self._author}/{self._repo}', self._repo, ''.join(x.title() for x in self._repo.split('-')), framework]
        for file in Path(file_temp).glob('*'):
            if file.name.endswith(('.md', '.yaml', 'template.py')):
                new_file = Path(file_dest) / str(file.name).replace(repo_temp.replace('-', '_'), self._repo.replace('-', '_'))
                self.hub_utils.update_text(ori_str_list, tar_str_list, str(file), str(new_file))
            elif file.name != '.git':
                os.rename(file, Path(file_dest) / file.name)
        if framework != 'pytorch':
            os.rename(Path(file_dest) / 'pytorch', Path(file_dest) / framework)

    def init_pyoperator(self, file_temp: Union[str, Path], file_dest: Union[str, Path]) -> None:
        """
        Initialize the files under file_dest by moving and updating the text under file_temp.

        Args:
            file_temp (`Union[str, Path]`):
                The path to the template files.
            file_dest (`Union[str, Path]`):
                The path to the local repo to init.

        Raises:
            (`HTTPError`)
                Raise error in request.
            (`OSError`)
                Raise error in writing file.
        """
        repo_temp = self._temp['pyoperator']

        ori_str_list = [f'author/{repo_temp}', repo_temp, ''.join(x.title() for x in repo_temp.split('-'))]
        tar_str_list = [f'{self._author}/{self._repo}', self._repo, ''.join(x.title() for x in self._repo.split('-'))]
        for file in Path(file_temp).glob('*'):
            if file.name.endswith(('.md', '.yaml', 'template.py')):
                new_file = Path(file_dest) / str(file.name).replace(repo_temp.replace('-', '_'), self._repo.replace('-', '_'))
                self.hub_utils.update_text(ori_str_list, tar_str_list, str(file), str(new_file))
            elif file.name != '.git':
                os.rename(file, Path(file_dest) / file.name)

    def generate_yaml(self, local_dir: Union[str, Path] = Path.cwd()) -> None:
        """
        Generate the yaml of Operator.

        Args:
            local_dir (`Union[str, Path]`):
                The directory to the repo.
        """
        sys.path.append(str(local_dir))

        yaml_file = Path(local_dir) / (self._repo.replace('-', '_') + '.yaml')
        if yaml_file.exists():
            engine_log.error('There already exists %s.', yaml_file)
            return

        class_name = ''.join(x.title() for x in self._repo.split('-'))
        author_operator = self._author + '/' + self._repo

        # import the class from repo
        cls = getattr(import_module(self._repo.replace('-', '_')), class_name)
        init_args = cls.__init__.__annotations__
        try:
            del init_args['return']
        except KeyError:
            pass
        call_func = cls.__call__.__annotations__
        try:
            call_output = call_func.pop('return')
            call_output = call_output.__annotations__
        except KeyError:
            pass

        data = {
            'name': self._repo,
            'labels': {
                'recommended_framework': '', 'class': '', 'others': ''
            },
            'operator': author_operator,
            'init': self.hub_utils.convert_dict(init_args),
            'call': {
                'input': self.hub_utils.convert_dict(call_func), 'output': self.hub_utils.convert_dict(call_output)
            }
        }
        with open(yaml_file, 'a', encoding='utf-8') as outfile:
            dump_yaml(data, outfile)

    def check(self, local_dir: Union[str, Path] = Path().cwd()) -> bool:
        """
        Check if the main file exists and match the file name.

        Args:
            local_dir (`Union[str, Path]`):
                The directory to the repo.

        Returns:
            (`bool`)
                Check if passed.
        """
        # check if the main file exists and match the file name
        file_name = self._repo.replace('-', '_')
        for file in [f'{file_name}.py', f'{file_name}.yaml']:
            if not (Path(local_dir) / file).exists():
                return False
        return self.check_yaml(local_dir)

    def check_yaml(self, local_dir: Union[str, Path] = Path().cwd()) -> bool:
        """
        Check if the yaml file matches the format.

        Args:
            local_dir (`Union[str, Path]`):
                The directory to the repo.

        Returns:
            (`bool`)
                Check if passed.
        """
        try:
            yaml_file = Path(local_dir) / (self._repo.replace('-', '_') + '.yaml')
            with open(yaml_file, 'r', encoding='utf-8') as input_file:
                dicts = load_yaml(input_file)
                if 'init' in dicts.keys() and dicts['call']['input'] is not None and dicts['call']['output'] is not None:
                    return True
        except KeyError:
            return False



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from enum import Enum, auto
from collections import defaultdict

from towhee.dag.operator_repr import OperatorRepr
from towhee.engine.operator_runner.runner_base import RunnerStatus
from towhee.dataframe import DataFrame
from towhee.engine.operator_io import create_reader, create_writer
from towhee.engine.operator_runner import create_runner
from towhee.engine.thread_pool_task_executor import ThreadPoolTaskExecutor


class OpStatus(Enum):
    NOT_RUNNING = auto()
    RUNNING = auto()
    FINISHED = auto()
    FAILED = auto()
    STOPPED = auto()


class OperatorContext:
    """
    The OperatorContext manages an operator's input data and output data at runtime,
    as well as the operators' dependency within a GraphContext.
    The abstraction of OperatorContext hides the complexity of Dataframe management,
    input iteration, and data dependency between Operators. It offers a Task-based
    scheduling context.
    Args:
        op_repr: (OperatorRepr)
            The operator representation
        dataframes: (`dict` of `DataFrame`)
            All the `DataFrames` in `GraphContext`
    """

    def __init__(self, op_repr: OperatorRepr, dataframes: Dict[str, DataFrame]):
        self._repr = op_repr
        self._readers = OperatorContext._create_reader(op_repr, dataframes)
        self._writer = OperatorContext._create_writer(op_repr, dataframes)
        self._op_runners = []
        self._op_status = OpStatus.NOT_RUNNING
        self._err_msg = None

    @staticmethod
    def _create_reader(op_repr, dataframes):
        inputs_index = defaultdict(dict)
        for item in op_repr.inputs:
            inputs_index[item['df']][item['name']] = item['col']
        iter_type = op_repr.iter_info['type']
        iter_params = op_repr.iter_info.get('params')

        inputs = dict((item['df'], dataframes[item['df']]) for item in op_repr.inputs)
        readers = []
        for df_name, indexs in inputs_index.items():
            readers.append(create_reader(inputs[df_name], iter_type, indexs, iter_params))
        return readers

    @staticmethod
    def _create_writer(op_repr, dataframes):
        '''
        Normally one op one output dataframe.
        '''
        outputs = list({dataframes[output['df']] for output in op_repr.outputs})
        iter_type = op_repr.iter_info['type']
        return create_writer(iter_type, outputs)

    @property
    def name(self):
        return self._repr.name

    @property
    def err_msg(self):
        return self._err_msg

    @property
    def status(self):
        """
        Calc op-ctx status by checking all runners of this op-ctx
        """
        if self._op_status in [OpStatus.FINISHED, OpStatus.FAILED]:
            return self._op_status

        if len(self._op_runners) == 0:
            return self._op_status

        finished_count = 0
        for runner in self._op_runners:
            if runner.status == RunnerStatus.FAILED:
                self._op_status = OpStatus.FAILED
                self._err_msg = runner.msg
            else:
                if runner.status == RunnerStatus.FINISHED:
                    finished_count += 1
        if finished_count == len(self._op_runners):
            self._op_status = OpStatus.FINISHED
        return self._op_status

    def start(self, executor: ThreadPoolTaskExecutor) -> None:
        if self._op_status != OpStatus.NOT_RUNNING:
            raise RuntimeError('OperatorContext can only be started once')

        self._op_status = OpStatus.RUNNING

        try:
            for i in range(self._repr.threads):
                self._op_runners.append(
                    create_runner(
                        self._repr.iter_info['type'],
                        self._repr.name,
                        i,
                        self._repr.name,
                        self._repr.tag,
                        self._repr.function,
                        self._repr.init_args,
                        self._readers,
                        self._writer,
                    )
                )
        except AttributeError as e:
            self._err_msg = str(e)
            self._op_status = OpStatus.FAILED
            return

        for runner in self._op_runners:
            executor.push_task(runner)

    def slow_down(self, time_sec: int):
        if self.status == OpStatus.RUNNING:
            for runner in self._op_runners:
                runner.slow_down(time_sec)

    def speed_up(self):
        if self.status == OpStatus.RUNNING:
            for runner in self._op_runners:
                runner.speed_up()

    def stop(self):
        if self.status != OpStatus.RUNNING:
            raise RuntimeError('Op ctx is already not running.')

        for runner in self._op_runners:
            runner.set_stop()

    def join(self):
        # Wait all runners finished.
        for runner in self._op_runners:
            runner.join()
        self._writer.close()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from towhee.engine.operator_runner.runner_base import RunnerBase


class ConcatRunner(RunnerBase):
    """
    Concat wrapper, multiple inputs one output.

    Concat can not run in multi-thread.
    """
    def __init__(
        self,
        name: str,
        index: int,
        op_name: str,
        tag: str,
        hub_op_id: str,
        op_args: Dict[str, any],
        readers=None,
        writer=None,
    ) -> None:
        if len(readers) <= 1:
            self._set_failed('Concat operator\'s inputs are at leaset two, current is %s' % len(readers))
            return

        super().__init__(name, index, op_name, tag, hub_op_id, op_args, readers, writer)

    def process_step(self):
        self._op._readers = self._readers  # pylint: disable=protected-access
        self._op._writer = self._writer  # pylint: disable=protected-access
        self._op()
        self._set_finished()
        return True



# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.


def window_reverse(windows, window_size, b, d, h, w):
    """
    Args:
        windows (`torch.Tensor`):
            Tensor with size (B*num_windows, window_size, window_size, C)
        window_size (`tuple[int]`):
            3D window size.
        b (`int`):
            Batch size
        d (`int`):
            Window size in time dimension.
        h (`int`):
            Height of image
        w (`int`):
            Width of image
    Returns:
        x (`torch.Tensor`):
            Tensor with size (b, d, h, w, c)
    """
    x = windows.view(b, d // window_size[0], h // window_size[1], w // window_size[2],
                     window_size[0], window_size[1], window_size[2], -1)
    x = x.permute(0, 1, 4, 2, 5, 3, 6, 7).contiguous().view(b, d, h, w, -1)
    return x




# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from towhee.engine.operator_io import reader as io_reader
from towhee.engine.operator_io import writer as io_writer
from towhee.dataframe import DataFrame


def create_reader(
    df: DataFrame,
    iter_type: str,
    inputs_index: Dict[str, int],
    iter_params: Optional[Dict] = None
) -> io_reader.DataFrameReader:
    if iter_type.lower() in ['map', 'flatmap', 'concat', 'filter', 'generator']:
        return io_reader.BlockMapReaderWithOriginData(df, inputs_index)
    elif iter_type.lower() == 'window':
        return io_reader.BatchFrameReader(df, inputs_index, **iter_params)
    elif iter_type.lower() == 'time_window':
        return io_reader.TimeWindowReader(df, inputs_index, **iter_params)
    else:
        raise NameError('Can not find %s iters' % iter_type)


def create_writer(iter_type: str, outputs: List[DataFrame]) -> io_writer.DataFrameWriter:
    assert len(outputs) == 1
    if iter_type.lower() in ['map', 'flatmap', 'concat', 'window', 'generator', 'time_window']:
        return io_writer.DictDataFrameWriter(outputs[0])
    elif iter_type.lower() == 'filter':
        return io_writer.RowDataFrameWriter(outputs[0])
    else:
        raise NameError('Can not find %s iters' % iter_type)



from typing import Union


class LayerFreezer:
    """
    Utilities to freeze/unfreeze layers.

    Args:
        model:
            a model with weights.
    Return:
        LayerFreezer.

    Example:
        >>> from towhee.trainer.utils.layer_freezer import LayerFreezer
        >>> from towhee.models.vit.vit import VitModel
        >>> my_model = VitModel('vit_base_patch16_224')
        >>> my_freezer = LayerFreezer(my_model)
        >>> # Check if modules in the last layer are frozen
        >>> my_freezer.status(-1)
        >>> # Check if modules in the layer "head" are frozen
        >>> my_freezer.status("head")
        ['unfrozen', 'unfrozen']
        >>> # Show all frozen layers
        >>> my_freezer.show_frozen_layers()
        ['patch_embed', 'head']
        >>> # Freeze layers by a list of layer indexes
        >>> my_freezer.by_idx([0, -1])
        >>> # Freeze layers by a list of layer names
        >>> my_freezer.by_names(['head'])
        >>> # Freeze all layers
        >>> my_freezer.set_all()
        >>> # Unfreeze all layers
        >>> my_freezer.set_all(freeze=False)
        >>> # Freeze all except the last layer
        >>> my_freezer.set_slice(-1)
    """

    def __init__(self, model):
        self.model = model
        self.layer_names = list(dict(self.model.named_children()).keys())
        self.layer_count = len(self.layer_names)

    def status(self, layer: Union[str, int]):
        """
        Check if a layer is frozen or not by its name or index

        Args:
            layer (`Union[str, int]`):
                the name or index of layer.
        Return:
            A list of status ('frozen' or 'unfrozen') to indicate if modules in the layer are frozen or not.
        """
        if isinstance(layer, str):
            if layer not in self.layer_names:
                raise ValueError(f'Layer does not exist: {layer}')
            for name, child in self.model.named_children():
                if name == layer:
                    outs = []
                    for param in child.parameters():
                        outs.append('unfrozen' if param.requires_grad else 'frozen')
                    print(f'Freeze status of layer {layer}: {outs}')
                    return outs
        if isinstance(layer, int):
            outs = []
            for param in list(self.model.children())[layer].parameters():
                outs.append('unfrozen' if param.requires_grad else 'frozen')
            print(f'Freeze status of layer {layer}: {outs}')
            return outs

    def show_frozen_layers(self):
        """
        Show all names of frozen layers

        Args:
            None.
        Return:
            A list of names of frozen layers
        """
        outs = []
        for name, layer in self.model.named_children():
            flag = 0
            for param in layer.parameters():
                if not param.requires_grad:
                    flag = 1
            if flag == 1:
                outs.append(name)
        # print(f'{len(outs)} frozen layers (names): {outs}')
        return outs

    def by_names(self, names: list, freeze: bool = True):
        """
        Freeze/unfreeze layers by names

        Args:
            names (`list`):
                a list of layer names
            freeze (`bool`):
                if or not freeze layers (default: True)
        """
        if not set(names).issubset(set(self.layer_names)):
            invalid_names = set(names) - set(self.layer_names)
            raise ValueError(f'Layer does not exist: {invalid_names}')
        for name, layer in self.model.named_children():
            if name not in names:
                pass
            else:
                for param in layer.parameters():
                    param.requires_grad = not freeze

    def by_idx(self, idx: list, freeze: bool = True):
        """
        Freeze/unfreeze layers by indexes

        Args:
            idx (`list`):
                a list of layer indexes
            freeze (`bool`):
                if or not freeze layers (default: True)
        """
        for i in idx:
            for param in list(self.model.children())[i].parameters():
                param.requires_grad = not freeze

    def set_all(self, freeze: bool = True):
        """
        Freeze/unfreeze all layers.

        Args:
            freeze (`bool`):
                if or not freeze layers (default: True)
        """
        for layer in self.model.children():
            for param in layer.parameters():
                param.requires_grad = not freeze

    def set_slice(self, slice_num: int, freeze: bool = True):
        """
        Freeze/unfreeze layers by list slice.

        Args:
            slice_num (`int`):
                number to slice the list of layers
            freeze (`bool`):
                if or not freeze layers (default: True)
        """
        myslice = slice(slice_num)
        slice_idx = list(range(self.layer_count))[myslice]
        self.by_idx(idx=slice_idx, freeze=freeze)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import numpy as np


class Image:
    """
    The unified form of images defined in Towhee.

    Agrs:
        image (`bytes`):
            The bytes of the image.
        width (`int`):
            The width of the image.
        height (`int`):
            The height of the image.
        channel (`int`):
            The channel of the image.
        mode (`str`):
            The mode of the image(i.e. 'RGB', 'RGBA', 'HSV', etc.).
        array (`np.ndarray`):
            The image in the form of ndarray.
    """

    def __init__(self, image: bytes, width: int, height: int, channel: int, mode: str, array: np.ndarray = None, key_frame: bool = False):
        self._image = image
        self._width = width
        self._height = height
        self._channel = channel
        self._mode = mode
        self._array = array
        self._key_frame = key_frame

    @property
    def image(self) -> bytes:
        return self._image

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def channel(self) -> int:
        return self._channel

    @property
    def mode(self) -> str:
        return self._mode

    @property
    def key_frame(self):
        return self._key_frame

    @property
    def array(self) -> np.ndarray:
        if not isinstance(self._array, np.ndarray):
            raise AttributeError('The array of image is not given, please call `Image.to_ndarray()` function to get the ndarray.')
        else:
            return copy.deepcopy(self._array)

    def to_ndarray(self) -> np.ndarray:
        """
        Load the np.ndarray form of the image.
        """
        if not isinstance(self._array, np.ndarray):
            shape = (self._height, self._width, self._channel)
            self._array = np.ndarray(shape, np.uint8, self._image)
            # self._array = np.frombuffer(self._image, dtype=np.uint8).reshape(self._height, self._width, self._channel)

        return copy.deepcopy(self._array)



# Original paper:
#
#     "Time2Vec: Learning a Vector Representation of Time" https://arxiv.org/abs/1907.05321
#
# Implemented with Pytorch by Copyright 2022 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from torch import nn


class Time2Vec(nn.Module):
    """
    Time2Vec implementaion in Pytorch.

    Args:
        seq_len (`int`):
            the length of input sequence
        activation (`str`):
            activation functions used for periodic time embedding (only support "sin" or "cos")

    Return:
        embedding with the same shape as input

    Example:
        >>> from towhee.models.layers.time2vec import Time2Vec
        >>> import torch
        >>>
        >>> x = torch.randn(3, 64)
        >>> model = Time2Vec(seq_len=64, activation="sin")
        >>> model(x).shape
        torch.Size([3, 64])
    """
    def __init__(self, seq_len: int, activation: str, **kwargs):
        super().__init__(**kwargs)
        self.w0 = nn.parameter.Parameter(torch.randn(seq_len, 1))
        self.b0 = nn.parameter.Parameter(torch.randn(1))
        self.w = nn.parameter.Parameter(torch.randn(seq_len, 1))
        self.b = nn.parameter.Parameter(torch.randn(1))

        if activation == 'sin':
            self.f = torch.sin
        elif activation == 'cos':
            self.f = torch.cos
        else:
            raise ValueError(f'Activation {activation} is not supported yet.')

        self.fc = nn.Linear(2, seq_len)

    def forward(self, x):
        periodic = self.f(torch.matmul(x, self.w) + self.b)  # Periodic time embedding
        linear = torch.matmul(x, self.w0) + self.b0  # Linear time embedding
        out = torch.cat([periodic, linear], -1)  # Concat time embeddings
        out = self.fc(out)  # Convert embedding dimension back to sequence length
        return out



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from towhee.utils.log import engine_log


class _Schema:
    """
    Schema of dataframe.
    Every dataframe has _Frame cols.
    """

    def __init__(self):
        self._cols = []
        self._key_index = {}
        self._sealed = False

    def add_col(self, name: str, col_type: str) -> bool:
        if self._sealed:
            engine_log.error('Schema already sealed, can not change cols')
            return False

        if name in self._key_index:
            engine_log.error('Col name {%s} already exist', name)
            return False

        self._cols.append((name, col_type))
        self._key_index[name] = len(self._cols) - 1
        return True

    def col_index(self, col_name: str) -> int:
        return self._key_index[col_name]

    def col_key(self, index: int) -> str:
        return self._cols[index][0]

    def col_type(self, index: int):
        return self._cols[index][1]

    def seal(self):
        self._sealed = True

    @property
    def col_count(self):
        return len(self._cols)

    @property
    def cols(self):
        return self._cols




# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
from typing import Union, Dict, List

from ruamel.yaml import YAML


def load_yaml(stream, typ: str = 'safe'):
    """
    Load the YAML document in a stream as Python object.

    Args:
        stream:
            The YAML file loaded in a stream.
        typ:
            The type of Dumper/Loader to apply in ruamel.yaml.

    Returns:
        The YAML laoded as python object.
    """
    yaml = YAML(typ=typ)
    data = yaml.load(stream=stream)
    if not isinstance(data, dict):
        raise ValueError(
            'The loaded data should be a dict, please check your yaml source.\
            (Such error is very likely caused by using dash instead of underline in yaml file name.)'
        )

    return data


def dump_yaml(data: Union[str, Dict, List], stream=None):
    """
    Dump YAML file as python object.

    Args:
        data (`Union[str, Dict, List]`):
            The data to write into a YAMl file.
        stream:
            The stream to dump the YAML into.
    """
    yaml = YAML()
    yaml.indent(mapping=4, sequence=8, offset=4)
    yaml.compact_seq_map = False

    if stream is None:
        stream = io.StringIO()
        yaml.dump(data=data, stream=stream)
        return stream.getvalue()
    else:
        yaml.dump(data=data, stream=stream)



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn

from towhee.models.utils.init_vit_weights import init_vit_weights


class MultiscaleVisionTransformers(nn.Module):
    """
    Multiscale Vision Transformers
    https://arxiv.org/abs/2104.11227
    ::
                                       PatchEmbed
                                           ↓
                                   PositionalEncoding
                                           ↓
                                        Dropout
                                           ↓
                                     Normalization
                                           ↓
                                         Block 1
                                           ↓
                                           .
                                           .
                                           .
                                           ↓
                                         Block N
                                           ↓
                                     Normalization
                                           ↓
                                          Head
    Args:
        patch_embed ('nn.Module'):
            Patch embed module.
        cls_positional_encoding ('nn.Module'):
            Positional encoding module.
        pos_drop ('nn.Module'):
            Dropout module after patch embed.
        norm_patch_embed ('nn.Module'):
            Normalization module after patch embed.
        blocks ('nn.ModuleList'):
            Stack of multi-scale transformer blocks.
        norm_embed ('nn.Module'):
            Normalization layer before head.
        head ('nn.Module'):
            Head module.
    """

    def __init__(
        self,
        *,
        patch_embed,
        cls_positional_encoding,
        pos_drop,
        norm_patch_embed,
        blocks,
        norm_embed,
        head,
    ) -> None:
        super().__init__()
        self.patch_embed: nn.Module = patch_embed
        self.cls_positional_encoding: nn.Module = cls_positional_encoding
        self.pos_drop: nn.Module = pos_drop
        self.norm_patch_embed: nn.ModuleList = norm_patch_embed
        self.blocks: nn.ModuleList = blocks
        self.norm_embed: nn.Module = norm_embed
        self.head: nn.Module = head

        assert hasattr(
            cls_positional_encoding, "patch_embed_shape"
        ), "cls_positional_encoding should have attribute patch_embed_shape."
        init_vit_weights(self, trunc_normal_std=0.02)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.patch_embed is not None:
            x = self.patch_embed(x)
        x = self.cls_positional_encoding(x)

        if self.pos_drop is not None:
            x = self.pos_drop(x)

        if self.norm_patch_embed is not None:
            x = self.norm_patch_embed(x)

        thw = self.cls_positional_encoding.patch_embed_shape
        for blk in self.blocks:
            x, thw = blk(x, thw)
        if self.norm_embed is not None:
            x = self.norm_embed(x)
        if self.head is not None:
            x = self.head(x)
        return x



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn

from towhee.models.layers.sequence_pool import SequencePool


class VisionTransformerBasicHead(nn.Module):
    """
    Vision transformer basic head.
    ::
                                      SequencePool
                                           ↓
                                        Dropout
                                           ↓
                                       Projection
                                           ↓
                                       Activation
    Args:
        in_features ('int'):
            Input channel size of the resnet head.
        out_features ('int'):
            Output channel size of the resnet head.
        seq_pool_type ('str'):
            Pooling type. It supports "cls", "mean " and "none". If set to
            "cls", it assumes the first element in the input is the cls token and
            returns it. If set to "mean", it returns the mean of the entire sequence.
        activation ('callable'):
            A callable that constructs vision transformer head activation layer,
            examples include: nn.ReLU, nn.Softmax, nn.Sigmoid, and None (not applying activation).
        dropout_rate ('float'):
            Dropout rate.
    """

    def __init__(
        self,
        *,
        in_features,
        out_features,
        seq_pool_type="cls",
        dropout_rate=0.5,
        activation=None,
    ) -> None:
        super().__init__()
        assert seq_pool_type in ["cls", "mean", "none"]

        if seq_pool_type in ["cls", "mean"]:
            self.seq_pool_model = SequencePool(seq_pool_type)
        elif seq_pool_type == "none":
            self.seq_pool_model = None
        else:
            raise NotImplementedError

        if activation is None:
            self.activation_model = None
        elif activation == nn.Softmax:
            self.activation_model = self.activation(dim=1)
        else:
            self.activation_model = self.activation()
        self.dropout = nn.Dropout(dropout_rate) if dropout_rate > 0.0 else None
        self.proj = nn.Linear(in_features, out_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Performs pooling.
        if self.seq_pool_model is not None:
            x = self.seq_pool_model(x)
        # Performs dropout.
        if self.dropout is not None:
            x = self.dropout(x)
        # Performs projection.
        x = self.proj(x)
        # Performs activation.
        if self.activation_model is not None:
            x = self.activation_model(x)
        return x



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.hparam import param_scope

class ExecuteCommand:
    """
    Implementation for subcmd `towhee execute`
    """
    def __init__(self, args) -> None:
        self._args = args

    def __call__(self):
        with param_scope(*self._args.define) as hp:
            import sys        # pylint: disable=import-outside-toplevel
            sys.path.append('.')
            import threading  # pylint: disable=import-outside-toplevel
            threading.setprofile(lambda f, e, a: param_scope.init(hp))

            if self._args.dry_run:
                hp().towhee.dry_run = True
            from towhee import pipeline  # pylint: disable=import-outside-toplevel
            pipe = pipeline(self._args.pipeline)

            if self._args.dry_run:
                print(repr(pipe))
            else:
                i = 0
                while i < self._args.iterations:
                    output = pipe(0)[0][0]

                    if self._args.output == 'imshow':
                        from towhee.utils.cv2_utils import cv2 # pylint: disable=import-outside-toplevel
                        cv2.imshow('imshow', output)
                        cv2.waitKey(1)
                    i += 1


    @staticmethod
    def install(subparsers):
        parser_execute = subparsers.add_parser('execute',
                                               help='execute pipeline')
        parser_execute.add_argument('-d',
                                    '--dry_run',
                                    action='store_true',
                                    help='dry run')
        parser_execute.add_argument('-n', '--iterations', type=int, default=-1)
        parser_execute.add_argument('-o',
                                    '--output',
                                    type=str,
                                    default='imshow')
        parser_execute.add_argument('-D', '--define', nargs='*', default=[])
        parser_execute.add_argument('pipeline', type=str, help='pipeline uri')



# Copyright lyakaap. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

import torch
import torch.nn.functional as F

from torch import nn


class NetVLAD(nn.Module):
    """
    NetVLAD layer implementation.

    Args:
            num_clusters (`int`): The number of clusters
            dim (`int`): Dimension of descriptors
            alpha (`float`): Parameter of initialization. Larger value is harder assignment.
            normalize_input (`bool`): If true, descriptor-wise L2 normalization is applied to input.
    """

    def __init__(self, num_clusters: int = 64, dim: int = 128, alpha: float = 100.0,
                 normalize_input: bool = True):
        super().__init__()
        self.num_clusters = num_clusters
        self.alpha = alpha
        self.normalize_input = normalize_input
        self.conv = nn.Conv2d(dim, num_clusters, kernel_size=(1, 1), bias=True)
        self.centroids = nn.Parameter(torch.rand(num_clusters, dim))
        self._init_params()

    def _init_params(self):
        self.conv.weight = nn.Parameter(
            (2.0 * self.alpha * self.centroids).unsqueeze(-1).unsqueeze(-1)
        )
        self.conv.bias = nn.Parameter(
            - self.alpha * self.centroids.norm(dim=1)
        )

    def forward(self, x: torch.Tensor):
        num_sample, in_dim = x.shape[:2]

        if self.normalize_input:
            x = F.normalize(x, p=2, dim=1)  # across descriptor dim

        # soft-assignment
        soft_assign = self.conv(x).view(num_sample, self.num_clusters, -1)
        soft_assign = F.softmax(soft_assign, dim=1)

        x_flatten = x.view(num_sample, in_dim, -1)

        # calculate residuals to each clusters
        residual = x_flatten.expand(self.num_clusters, -1, -1, -1).permute(1, 0, 2, 3) - \
                   self.centroids.expand(x_flatten.size(-1), -1, -1).permute(1, 2, 0).unsqueeze(0)
        residual *= soft_assign.unsqueeze(2)
        vlad = residual.sum(dim=-1)

        vlad = F.normalize(vlad, p=2, dim=2)  # intra-normalization
        vlad = vlad.view(x.size(0), -1)  # flatten
        vlad = F.normalize(vlad, p=2, dim=1)  # L2 normalize

        return vlad


class EmbedNet(nn.Module):
    """
    Embed a base model and the net vlad to a new network.

    Args:
            base_model (`int`): The base model that extracts image features
            net_vlad (`int`): NetVLAD model to extract global features
    """
    def __init__(self, base_model: torch.nn.Sequential, net_vlad: NetVLAD):
        super().__init__()
        self.base_model = base_model
        self.net_vlad = net_vlad

    def forward(self, x):
        x = self.base_model(x)
        embedded_x = self.net_vlad(x)
        return embedded_x



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Union, Dict, List
import traceback

from towhee.engine.operator_runner.runner_base import RunnerBase
from towhee.engine.status import Status
from towhee.errors import OpIOTypeError


class WindowRunner(RunnerBase):
    """
    WindowRunner, multiple inputs and multiple outputs.

    Wrapper will run in task executor in another thread or process.
    If run an op error, we should pass error info by an error handler.
    """

    def _get_inputs(self) -> Tuple[bool, Dict[str, any]]:
        try:
            data = self._reader.read()
            return False, data
        except StopIteration:
            return True, None

    def _set_outputs(self, output: List[any]):
        if not isinstance(output, list):
            raise OpIOTypeError("Window operator's output must be a list, not a {}".format(type(output)))

        for data in output:
            self._writer.write(data._asdict())

    def _call_op(self, inputs: List[Dict]) -> Tuple[bool, Union[str, any]]:
        try:
            outputs = self._op(inputs)
            return Status.ok_status(outputs)
        except Exception as e:  # pylint: disable=broad-except
            err = "{}, {}".format(e, traceback.format_exc())
            return Status.err_status(err)



# Copyright 2021 Microsoft . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
import torch
from torch import nn
from torch.utils import checkpoint
from .swin_transformer_block import SwinTransformerBlock

class BasicLayer(nn.Module):
    """ A basic Swin Transformer layer for one stage.
    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resolution.
        depth (int): Number of blocks.
        num_heads (int): Number of attention heads.
        window_size (int): Local window size.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float | tuple[float], optional): Stochastic depth rate. Default: 0.0
        norm_layer (nn.Module, optional): Normalization layer. Default: nn.LayerNorm
        downsample (nn.Module | None, optional): Downsample layer at the end of the layer. Default: None
        use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False.
    """

    def __init__(self, dim, input_resolution, depth, num_heads, window_size,
                 mlp_ratio=4., qkv_bias=True, drop=0., attn_drop=0.,
                 drop_path=0., norm_layer=nn.LayerNorm, downsample=None, use_checkpoint=False):

        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.depth = depth
        self.use_checkpoint = use_checkpoint

        # build blocks
        self.blocks = nn.ModuleList([
            SwinTransformerBlock(
                dim=dim, input_resolution=input_resolution, num_heads=num_heads, window_size=window_size,
                shift_size=0 if (i % 2 == 0) else window_size // 2, mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias, drop=drop, attn_drop=attn_drop,
                drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path, norm_layer=norm_layer)
            for i in range(depth)])

        # patch merging layer
        if downsample is not None:
            self.downsample = downsample(input_resolution, dim=dim, norm_layer=norm_layer)
        else:
            self.downsample = None

    def forward(self, x):
        for blk in self.blocks:
            if not torch.jit.is_scripting() and self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x)
            else:
                x = blk(x)
        if self.downsample is not None:
            x = self.downsample(x)
        return x

    def extra_repr(self) -> str:
        return f'dim={self.dim}, input_resolution={self.input_resolution}, depth={self.depth}'



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
from dataclasses import dataclass
from typing import Union, Dict, Any, Optional, List

# from towhee.utils.log import trainer_log
from towhee.trainer.training_config import TrainingConfig

MODEL_CARD_NAME = "README.md"


@dataclass
class ModelCard:
    """
    Utilities to generate and save model card. Recommended attributes from https://arxiv.org/abs/1810.03993 (see papers)
    https://catalog.ngc.nvidia.com/orgs/nvidia/teams/tao/models/actionrecognitionnet

    Args:
        model_name (`Optional[str]`):
            model name
        model_architecture (`Optional[str]`):
            model structure
        model_overview (`Optional[str] = None`):
        language (`Optional[Union[str, List[str]]]`):
            language
        tags (`Optional[Union[str, List[str]]]`):
            tags
        tasks (`Optional[Union[str, List[str]]]`):
            model tasks (eg. classification, prediction, etc.)
        datasets (`Optional[Union[str, List[str]]]`):
            datasets used to train/test the model
        datasets_tags (`Optional[Union[str, List[str]]]`):
            tags of datasets
        dataset_args (`Optional[Union[str, List[str]]]`):
            arguments of dataset
        eval_results (`Optional[Dict[str, float]]`):
            evaluation results recorded
        eval_lines (`Optional[List[str]]`):
            evaluation baselines
        training_summary (`Optional[Dict[str, Any]]`):
            training summary include training information
        training_config (`Optional[TrainingConfig]`):
            training configurations
        source (`Optional[str]`):
            source of model card (default = "trainer")

    Example:
        >>> from towhee.trainer.modelcard import ModelCard
        >>> model_card = ModelCard(model_name='test')
        >>> # Print out model name stored in model card
        >>> model_card.model_name
        'test'
        >>> # Save model card to "path/to/my_dir" as README.md
        >>> model_card.save_model_card('/path/to/my_dir')
        >>> # Save model card as "/path/to/my_dir/model_card.md"
        >>> model_card.save_model_card('/path/to/my_dir/model_card.md')
    """
    model_name: Optional[str] = None
    model_architecture: Optional[str] = None
    model_overview: Optional[str] = None
    language: Optional[Union[str, List[str]]] = None
    # license: Optional[str] = None
    tags: Optional[Union[str, List[str]]] = None
    tasks: Optional[Union[str, List[str]]] = None
    datasets: Optional[Union[str, List[str]]] = None
    datasets_tags: Optional[Union[str, List[str]]] = None
    dataset_args: Optional[Union[str, List[str]]] = None
    eval_results: Optional[Dict[str, float]] = None
    eval_lines: Optional[List[str]] = None
    training_summary: Optional[Dict[str, Any]] = None
    training_config: Optional[TrainingConfig] = None
    source: Optional[str] = "trainer"

    def __post_init__(self):
        pass

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self._to_json_string())

    def save_model_card(self, save_directory_or_file):
        """
        Write model card to the given filepath or directory

        Args:
            save_directory_or_file (`str`):
                file path or directory to write and save model card.
        """
        model_card = f"# {self.model_name}"
        model_card += f"\n\n## Model overview\n{self.model_overview}\n"
        model_card += f"\n\n### Model architecture\n{self.model_architecture}\n"

        model_card += "\n\n## Dataset"
        if self.datasets is None:
            model_card += "\nDataset unknown.\n"
        else:
            if isinstance(self.datasets, str):
                model_card += f"\nUsing the {self.datasets} as dataset.\n"
            elif isinstance(self.dataset, (tuple, list)) and len(self.datasets) == 1:
                model_card += f"\nUsing the {self.datasets} as dataset.\n"
            else:
                model_card += (
                        "\n, ".join(
                            [f"Using the {dataset}" for dataset in self.datasets[:-1]])
                        + f" and the {self.datasets[-1]} dataset.\n"
                )

        model_card += "\n\n## Training configurations\n"
        if self.training_config is not None:
            if self.training_config.optimizer is not None:
                if isinstance(self.training_config.optimizer, str):
                    model_card += f"\nOptimizer is  {self.training_config.optimizer}.\n"
                else:
                    model_card += "\n".join([f"- {name}: "
                                             f"{value}" for name, value in self.training_config.optimizer.items()])
                    model_card += "\n"
            if self.training_config.lr_scheduler_type is not None:
                model_card += f"\nThe scheduler type is {self.training_config.lr_scheduler_type}.\n"
            if self.training_config.warmup_ratio is not None:
                model_card += f"\nThe warmup_ratio is {self.training_config.warmup_ratio}.\n"
            if self.training_config.warmup_steps is not None:
                model_card += f"\nThe warmup_steps is {self.training_config.warmup_steps}.\n"
            if self.training_config.lr is not None:
                model_card += f"\nLearning rate is {self.training_config.lr}.\n"
            else:
                model_card += "\nLearning rate is needed.\n"
            # if self.training_config.weight_decay is not None:
            #     model_card += f"\nWeight decay is {self.training_config.weight_decay}."
            # if self.training_config.adam_beta1 is not None:
            #     model_card += f"\nBeta1 for AdamW optimizer is {self.training_config.adam_beta1}."
            # if self.training_config.adam_beta2 is not None:
            #     model_card += f"\nBeta2 for AdamW optimizer is {self.training_config.adam_beta2}."
            # if self.training_config.adam_epsilon is not None:
            #     model_card += f"\nEpsilon for AdamW optimizer is {self.training_config.adam_epsilon}."
            if self.training_config.metric is not None:
                model_card += f"\nMetric is {self.training_config.metric}.\n"
            #
            # if self.training_config.max_norm_grad is not None:
            #     model_card += f"\nMax gradient norm is {self.training_config.max_norm_grad}."
            if self.training_config.batch_size is not None:
                model_card += f"\nBatch size is {self.training_config.batch_size}.\n"
            if self.training_config.seed is not None:
                model_card += f"\nRandom seed is {self.training_config.seed}.\n"
            if self.training_config.epoch_num is not None:
                model_card += f"\nTraining epochs is {self.training_config.epoch_num}.\n"
            # if self.training_config.n_gpu is not None:
            #     model_card += f"\nNumber of GPUs is {self.training_config.n_gpu}.\n"
        else:
            model_card += "\nTraining configurations is needed.\n"

        model_card += "\n\n## Training summary\n"
        if self.training_summary is not None:
            model_card += "\n"
            model_card += "\n".join([f"- {name}: {value}" for name, value in self.training_summary.items()])
            model_card += "\n"
        else:
            model_card += "\nTraining summary is needed.\n"

        if Path.is_dir(Path(save_directory_or_file)):
            # If we save using the predefined names, we can load using `from_pretrained`
            output_model_card_file = Path(save_directory_or_file).joinpath(MODEL_CARD_NAME)
        else:
            output_model_card_file = save_directory_or_file

        with open(output_model_card_file, "w", encoding="utf-8") as f:
            f.write(model_card)

    def _to_json_file(self, json_file_path):
        """Save this instance to a json file."""
        pass

    def _to_json_string(self):
        """Serializes this instance to a JSON string."""
        pass

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        pass

    @staticmethod
    def load_from_file(file_path):
        pass




# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Iterable, Iterator, Callable
from random import random, sample, shuffle

from towhee.hparam import param_scope
from towhee.functional.option import Option, Some, Empty
from towhee.functional.mixins import AllMixins


class DataCollection(Iterable, AllMixins):
    """
    DataCollection is a pythonic computation and processing framework
    for unstructured data in machine learning and data science.
    It allows a data scientist or researcher to assemble a data processing pipeline,
    do his model work (embedding, transforming, or classification)
    and apply it to the business (search, recommendation, or shopping)
    with a method-chaining style API.

    Examples:

    1. Create a data collection from list or iterator:

    >>> dc = DataCollection([0, 1, 2, 3, 4])
    >>> dc = DataCollection(iter([0, 1, 2, 3, 4]))

    2. Chaining function invocations makes your code clean and fluent:

    >>> (
    ...    dc.map(lambda x: x+1)
    ...      .map(lambda x: x*2)
    ... ).to_list()
    [2, 4, 6, 8, 10]

    3. Multi-line closures are also supported via decorator syntax

    >>> dc = DataCollection([1,2,3,4])
    >>> @dc.map
    ... def add1(x):
    ...     return x+1
    >>> @add1.map
    ... def mul2(x):
    ...     return x *2
    >>> mul2.to_list()
    [4, 6, 8, 10]

    >>> dc = DataCollection([1,2,3,4])
    >>> @dc.filter
    ... def ge3(x):
    ...     return x>=3
    >>> ge3.to_list()
    [3, 4]

    `DataCollection` is designed to behave as a python list or iterator. Consider you are running
    the following code:

    .. code-block:: python
      :linenos:

      dc.map(stage1)
        .map(stage2)

    1. `iterator` and `stream mode`: When a `DataCollection` object is created from an iterator, it behaves as a python
    iterator and performs `stream-wise` data processing:

        a. `DataCollection` takes one element from the input and applies `stage1` and `stage2` sequentially ;
        b. Since DataCollection holds no data, indexing or shuffle is not supported;

    2. `list` and `unstream mode`: If a `DataCollection` object is created from a list, it will hold all the input values,
    and perform stage-wise computations:

        a. `stage2` will wait until all the calculations are done in `stage1`;
        b. A new DataCollection will be created to hold all the outputs for each stage. You can perform list operations on result DataCollection;

    """

    def __init__(self, iterable: Iterable) -> None:
        """Initializes a new DataCollection instance.

        Args:
            iterable (Iterable): input data
        """
        super().__init__()
        self._iterable = iterable

    def __iter__(self):
        return iter(self._iterable)

    def stream(arg):  # pylint: disable=no-self-argument
        """
        Create a stream data collection.

        Examples:

        1. Create a streamed data collection

        >>> dc = DataCollection.stream([0,1,2,3,4])
        >>> dc.is_stream
        True

        2. Convert a data collection to streamed version

        >>> dc = DataCollection([0, 1, 2, 3, 4])
        >>> dc.is_stream
        False

        >>> dc = dc.stream()
        >>> dc.is_stream
        True
        """
        # pylint: disable=protected-access
        if isinstance(arg, DataCollection):
            self = arg
            if self.is_stream:
                return self
            return DataCollection.stream(self._iterable)

        iterable = arg
        if not isinstance(iterable, Iterator):
            iterable = iter(iterable)
        return DataCollection(iterable)

    def unstream(arg):  # pylint: disable=no-self-argument
        """
        Create a unstream data collection.

        Examples:

        1. Create a unstream data collection

        >>> dc = DataCollection.unstream(iter(range(5)))
        >>> dc.is_stream
        False

        2. Convert a streamed data collection to unstream version

        >>> dc = DataCollection(iter(range(5)))
        >>> dc.is_stream
        True
        >>> dc = dc.unstream()
        >>> dc.is_stream
        False
        """
        # pylint: disable=protected-access
        if isinstance(arg, DataCollection):
            self = arg
            if not self.is_stream:
                return self
            return DataCollection.unstream(self._iterable)

        iterable = arg
        if isinstance(iterable, Iterator):
            iterable = list(iterable)
        return DataCollection(iterable)

    @property
    def is_stream(self):
        """
        Check whether the data collection is stream or unstream.

        Examples:

        >>> dc = DataCollection([0,1,2,3,4])
        >>> dc.is_stream
        False

        >>> result = dc.map(lambda x: x+1)
        >>> result.is_stream
        False
        >>> result._iterable
        [1, 2, 3, 4, 5]

        >>> dc = DataCollection(iter(range(5)))
        >>> dc.is_stream
        True

        >>> result = dc.map(lambda x: x+1)
        >>> result.is_stream
        True
        >>> isinstance(result._iterable, Iterable)
        True

        """
        return isinstance(self._iterable, Iterator)

    @property
    def _factory(self):
        """
        Factory method for data collection.

        This factory method has been wrapped into a `param_scope()` which contains parent infomations.
        """
        creator = DataCollection.stream if self.is_stream else DataCollection.unstream

        def wrapper(*arg, **kws):
            with param_scope() as hp:
                hp().data_collection.parent = self
                return creator(*arg, **kws)

        return wrapper

    def exception_safe(self):
        """
        Making the data collection exception-safe by warp elements with `Option`.

        Examples:

        1. Exception breaks pipeline execution:

        >>> dc = DataCollection.range(5)
        >>> dc.map(lambda x: x / (0 if x == 3 else 2)).to_list()
        Traceback (most recent call last):
        ZeroDivisionError: division by zero

        2. Exception-safe execution

        >>> dc.exception_safe().map(lambda x: x / (0 if x == 3 else 2)).to_list()
        [Some(0.0), Some(0.5), Some(1.0), Empty(), Some(2.0)]

        >>> dc.exception_safe().map(lambda x: x / (0 if x == 3 else 2)).filter(lambda x: x < 1.5).to_list()
        [Some(0.0), Some(0.5), Some(1.0), Empty()]

        >>> dc.exception_safe().map(lambda x: x / (0 if x == 3 else 2)).filter(lambda x: x < 1.5, drop_empty=True).to_list()
        [Some(0.0), Some(0.5), Some(1.0)]
        """
        result = map(lambda x: Some(x) if not isinstance(x, Option) else x, self._iterable)
        return self._factory(result)

    def safe(self):
        """
        Shortcut for `exception_safe`
        """
        return self.exception_safe()

    def select_from(self, other):
        """
        Select data from dc with list(self).

        Examples:

        >>> dc1 = DataCollection([0.8, 0.9, 8.1, 9.2])
        >>> dc2 = DataCollection([[1, 2, 0], [2, 3, 0]])

        >>> dc3 = dc2.select_from(dc1)
        >>> list(dc3)
        [[0.9, 8.1, 0.8], [8.1, 9.2, 0.8]]
        """
        def inner(x):
            if isinstance(x, Iterable):
                return [other[i] for i in x]
            return other[x]

        result = map(inner, self._iterable)
        return self._factory(result)

    def fill_empty(self, default: Any = None) -> 'DataCollection':
        """
        Unbox `Option` values and fill `Empty` with default values.

        Args:
            default (Any): default value to replace empty values;

        Returns:
            DataCollection: data collection with empty values filled with `default`;

        Examples:

        >>> dc = DataCollection.range(5)
        >>> dc.safe().map(lambda x: x / (0 if x == 3 else 2)).fill_empty(-1.0).to_list()
        [0.0, 0.5, 1.0, -1.0, 2.0]
        """
        result = map(lambda x: x.get() if isinstance(x, Some) else default, self._iterable)
        return self._factory(result)

    def drop_empty(self, callback: Callable = None) -> 'DataCollection':
        """
        Unbox `Option` values and drop `Empty`.

        Args:
            callback (Callable): handler for empty values;

        Returns:
            DataCollection: data collection that drops empty values;

        Examples:

        >>> dc = DataCollection.range(5)
        >>> dc.safe().map(lambda x: x / (0 if x == 3 else 2)).drop_empty().to_list()
        [0.0, 0.5, 1.0, 2.0]

        Get inputs that case exceptions:

        >>> exception_inputs = []
        >>> result = dc.safe().map(lambda x: x / (0 if x == 3 else 2)).drop_empty(lambda x: exception_inputs.append(x.get().value))
        >>> exception_inputs
        [3]
        """
        if callback is not None:

            def inner(data):
                for x in data:
                    if isinstance(x, Empty):
                        callback(x)
                    if isinstance(x, Some):
                        yield x.get()

            result = inner(self._iterable)
        else:

            def inner(data):
                for x in data:
                    if isinstance(x, Some):
                        yield x.get()

            result = inner(self._iterable)
        return self._factory(result)

    def map(self, *arg):
        """
        Apply operator to data collection.

        Args:
            *arg (Callable): functions/operators to apply to data collection;

        Returns:
            DataCollection: data collections that contains computation results;

        Examples:

        >>> dc = DataCollection([1,2,3,4])
        >>> dc.map(lambda x: x+1).map(lambda x: x*2).to_list()
        [4, 6, 8, 10]
        """

        # mmap
        if len(arg) > 1:
            return self.mmap(list(arg))

        unary_op = arg[0]

        # smap map for stateful operator
        if hasattr(unary_op, 'is_stateful') and unary_op.is_stateful:
            return self.smap(unary_op)

        # pmap
        if self.get_executor() is not None:
            return self.pmap(unary_op)

        # map
        def inner(x):
            if isinstance(x, Option):
                return x.map(unary_op)
            else:
                return unary_op(x)

        result = map(inner, self._iterable)
        return self._factory(result)

    def zip(self, *others) -> 'DataCollection':
        """
        Combine two data collections.

        Args:
            *others (DataCollection): other data collections;

        Returns:
            DataCollection: data collection with zipped values;

        Examples:

        >>> dc1 = DataCollection([1,2,3,4])
        >>> dc2 = DataCollection([1,2,3,4]).map(lambda x: x+1)
        >>> dc3 = dc1.zip(dc2)
        >>> list(dc3)
        [(1, 2), (2, 3), (3, 4), (4, 5)]
        """
        return self._factory(zip(self, *others))

    def filter(self, unary_op: Callable, drop_empty=False) -> 'DataCollection':
        """
        Filter data collection with `unary_op`.

        Args:
            unary_op (`Callable`):
                Callable to decide whether to filter the element;
            drop_empty (`bool`):
                Drop empty values. Defaults to False.

        Returns:
            DataCollection: filtered data collection
        """

        # return filter(unary_op, self)
        def inner(x):
            if isinstance(x, Option):
                if isinstance(x, Some):
                    return unary_op(x.get())
                return not drop_empty
            return unary_op(x)

        return self._factory(filter(inner, self._iterable))

    def sample(self, ratio=1.0) -> 'DataCollection':
        """
        Sample the data collection.

        Args:
            ratio (float): sample ratio;

        Returns:
            DataCollection: sampled data collection;

        Examples:

        >>> dc = DataCollection(range(10000))
        >>> result = dc.sample(0.1)
        >>> ratio = len(result.to_list()) / 10000.
        >>> 0.09 < ratio < 0.11
        True
        """
        return self._factory(filter(lambda _: random() < ratio, self))

    @staticmethod
    def range(*arg, **kws):
        """
        Generate data collection with ranged numbers.

        Examples:

        >>> DataCollection.range(5).to_list()
        [0, 1, 2, 3, 4]
        """
        return DataCollection(range(*arg, **kws))

    def batch(self, size, drop_tail=False) -> 'DataCollection':
        """
        Create small batches from data collections.

        Args:
            size (int): window size;
            drop_tail (bool): drop tailing windows that not full;

        Returns:
            DataCollection: data collection of batched windows;

        Examples:

        >>> dc = DataCollection(range(10))
        >>> [list(batch) for batch in dc.batch(2)]
        [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]

        >>> dc = DataCollection(range(10))
        >>> [list(batch) for batch in dc.batch(3)]
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

        >>> dc = DataCollection(range(10))
        >>> [list(batch) for batch in dc.batch(3, drop_tail=True)]
        [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
        """
        def inner():
            buff = []
            for ele in self._iterable:
                buff.append(ele)
                if len(buff) == size:
                    yield DataCollection.unstream(buff)
                    buff = []
            if not drop_tail and len(buff) > 0:
                yield DataCollection.unstream(buff)

        return self._factory(inner())

    def rolling(self, size: int, drop_head=True, drop_tail=True):
        """
        Create rolling windows from data collections.

        Args:
            size (int): window size;
            drop_head (bool): drop headding windows that not full;
            drop_tail (bool): drop tailing windows that not full;

        Returns:
            DataCollection: data collection of rolling windows;

        Examples:

        >>> dc = DataCollection(range(5))
        >>> [list(batch) for batch in dc.rolling(3)]
        [[0, 1, 2], [1, 2, 3], [2, 3, 4]]

        >>> dc = DataCollection(range(5))
        >>> [list(batch) for batch in dc.rolling(3, drop_head=False)]
        [[0], [0, 1], [0, 1, 2], [1, 2, 3], [2, 3, 4]]

        >>> dc = DataCollection(range(5))
        >>> [list(batch) for batch in dc.rolling(3, drop_tail=False)]
        [[0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4], [4]]
        """
        def inner():
            buff = []
            for ele in self._iterable:
                buff.append(ele)
                if not drop_head or len(buff) == size:
                    yield DataCollection.unstream(buff.copy())
                if len(buff) == size:
                    buff = buff[1:]
            while not drop_tail and len(buff) > 0:
                yield DataCollection.unstream(buff)
                buff = buff[1:]

        return self._factory(inner())

    def flatten(self) -> 'DataCollection':
        """
        Flatten nested data collections.

        Returns:
            DataCollection: flattened data collection;

        Examples:

        >>> dc = DataCollection(range(10))
        >>> nested_dc = dc.batch(2)
        >>> nested_dc.flatten().to_list()
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        """
        def inner():
            for ele in self._iterable:
                if isinstance(ele, Iterable):
                    for nested_ele in iter(ele):
                        yield nested_ele
                else:
                    yield ele

        return self._factory(inner())

    def shuffle(self, in_place=False) -> 'DataCollection':
        """
        Shuffle an unstreamed data collection.

        Args:
            in_place (bool): shuffle the data collection in place;

        Returns:
            DataCollection: shuffled data collection;

        Examples:

        1. Shuffle:

        >>> dc = DataCollection([0, 1, 2, 3, 4])
        >>> shuffled = dc.shuffle()
        >>> tuple(dc) == tuple(range(5))
        True

        >>> tuple(shuffled) == tuple(range(5))
        False

        2. In place shuffle:

        >>> dc = DataCollection([0, 1, 2, 3, 4])
        >>> _ = dc.shuffle(True)
        >>> tuple(dc) == tuple(range(5))
        False

        3. streamed data collection is not supported:

        >>> dc = DataCollection.stream([0, 1, 2, 3, 4])
        >>> _ = dc.shuffle()
        Traceback (most recent call last):
        TypeError: shuffle is not supported for streamed data collection.
        """
        if self.is_stream:
            raise TypeError('shuffle is not supported for streamed data collection.')
        if in_place:
            shuffle(self._iterable)
            return self
        else:
            return DataCollection.unstream(sample(self._iterable, len(self._iterable)))

    def __getattr__(self, name):
        """
        Unknown method dispatcher.

        When a unknown method is invoked on a `DataCollection` object,
        the function call will be dispatched to a method resolver.
        By registering function to the resolver, you are able to extend
        `DataCollection`'s API at runtime without modifying its code.

        Examples:

        1. Define two operators:

        >>> class my_add:
        ...     def __init__(self, val):
        ...         self.val = val
        ...     def __call__(self, x):
        ...         return x+self.val

        >>> class my_mul:
        ...     def __init__(self, val):
        ...         self.val = val
        ...     def __call__(self, x):
        ...         return x*self.val

        2. Register the operators to `DataCollection`'s execution context with `param_scope`:

        >>> from towhee import param_scope
        >>> with param_scope(dispatcher={
        ...         'add': my_add, # register `my_add` as `dc.add`
        ...         'mul': my_mul  # register `my_mul` as `dc.mul`
        ... }):
        ...     dc = DataCollection([1,2,3,4])
        ...     dc.add(1).mul(2).to_list() # call registered operator
        [4, 6, 8, 10]
        """
        with param_scope() as hp:
            dispatcher = hp().dispatcher({})

            def wrapper(path, index, *arg, **kws):
                _ = index
                op = self.resolve(dispatcher, path, index, *arg, **kws)
                return self.map(op)

            callholder = hp.callholder(wrapper)
        return getattr(callholder, name)

    def __getitem__(self, index):
        """
        Indexing for data collection.

        Examples:

        >>> dc = DataCollection([0, 1, 2, 3, 4])
        >>> dc[0]
        0

        >>> dc.stream()[1]
        Traceback (most recent call last):
        TypeError: indexing is not supported for streamed data collection.
        """
        if self.is_stream:
            raise TypeError('indexing is not supported for streamed data collection.')
        return self._iterable[index]

    def __setitem__(self, index, value):
        """
        Indexing for data collection.

        Examples:

        >>> dc = DataCollection([0, 1, 2, 3, 4])
        >>> dc[0]
        0

        >>> dc[0] = 5
        >>> dc._iterable[0]
        5

        >>> dc.stream()[0]
        Traceback (most recent call last):
        TypeError: indexing is not supported for streamed data collection.
        """
        if self.is_stream:
            raise TypeError('indexing is not supported for streamed data collection.')
        self._iterable[index] = value

    def __rshift__(self, unary_op):
        """
        Chain the operators with `>>`.

        Examples:

        >>> dc = DataCollection([1,2,3,4])
        >>> (dc
        ...     >> (lambda x: x+1)
        ...     >> (lambda x: x*2)
        ... ).to_list()
        [4, 6, 8, 10]
        """
        return self.map(unary_op)

    def __or__(self, unary_op):
        return self.map(unary_op)

    def __add__(self, other):
        """
        Concat two data collections:

        Examples:

        >>> (DataCollection.range(5) + DataCollection.range(5)).to_list()
        [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]

        >>> (DataCollection.range(5) + DataCollection.range(5) + DataCollection.range(5)).to_list()
        [0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 0, 1, 2, 3, 4]
        """
        def inner():
            for x in self:
                yield x
            for x in other:
                yield x

        return self._factory(inner())

    def head(self, n: int = 5):
        """
        Get the first n lines of a DataCollection.

        Args:
            n (`int`):
                The number of lines to print. Default value is 5.

        Examples:

        >>> DataCollection.range(10).head(3).to_list()
        [0, 1, 2]
        """
        def inner():
            for i, x in enumerate(self._iterable):
                if i >= n:
                    break
                yield x
        return self._factory(inner())

    def to_list(self):
        return self._iterable if isinstance(self._iterable, list) else list(self)


if __name__ == '__main__':  # pylint: disable=inconsistent-quotes
    import doctest

    doctest.testmod(verbose=False)



IMAGENET_DEFAULT_MEAN = [0.485, 0.456, 0.406]
IMAGENET_DEFAULT_STD = [0.229, 0.224, 0.225]


def _cfg(url='', **kwargs):
    return {
        'url': url,
        'num_classes': 1000, 'input_size': (3, 224, 224), 'pool_size': None,
        'crop_pct': .9, 'interpolation': 'bicubic', 'fixed_input_size': True,
        'mean': IMAGENET_DEFAULT_MEAN, 'std': IMAGENET_DEFAULT_STD,
        'first_conv': 'patch_embed.proj', 'classifier': 'head',
        **kwargs
    }

model_cfgs = {
    # patch models (my experiments)
    'swin_base_patch4_window12_384': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_base_patch4_window12_384_22kto1k.pth',
        input_size=(3, 384, 384), crop_pct=1.0),

    'swin_base_patch4_window7_224': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_base_patch4_window7_224_22kto1k.pth',
    ),

    'swin_large_patch4_window12_384': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_large_patch4_window12_384_22kto1k.pth',
        input_size=(3, 384, 384), crop_pct=1.0),

    'swin_large_patch4_window7_224': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_large_patch4_window7_224_22kto1k.pth',
    ),

    'swin_small_patch4_window7_224': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_small_patch4_window7_224.pth',
    ),

    'swin_tiny_patch4_window7_224': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_tiny_patch4_window7_224.pth',
    ),

    'swin_base_patch4_window12_384_in22k': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_base_patch4_window12_384_22k.pth',
        input_size=(3, 384, 384), crop_pct=1.0, num_classes=21841),

    'swin_base_patch4_window7_224_in22k': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_base_patch4_window7_224_22k.pth',
        num_classes=21841),

    'swin_large_patch4_window12_384_in22k': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_large_patch4_window12_384_22k.pth',
        input_size=(3, 384, 384), crop_pct=1.0, num_classes=21841),

    'swin_large_patch4_window7_224_in22k': _cfg(
        url='https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_large_patch4_window7_224_22k.pth',
        num_classes=21841),

}

def build_configs(name, **kwargs):

    config = model_cfgs[name]

    model_architectures = {
        'swin_base_patch4_window12_384' : dict(
            patch_size=4, window_size=12, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32), **kwargs),

        'swin_base_patch4_window7_224' : dict(
            patch_size=4, window_size=7, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32), **kwargs),

        'swin_large_patch4_window12_384' : dict(
            patch_size=4, window_size=12, embed_dim=192, depths=(2, 2, 18, 2), num_heads=(6, 12, 24, 48), **kwargs),

        'swin_large_patch4_window7_224' : dict(
            patch_size=4, window_size=7, embed_dim=192, depths=(2, 2, 18, 2), num_heads=(6, 12, 24, 48), **kwargs),

        'swin_small_patch4_window7_224' : dict(
            patch_size=4, window_size=7, embed_dim=96, depths=(2, 2, 18, 2), num_heads=(3, 6, 12, 24), **kwargs),

        'swin_tiny_patch4_window7_224' : dict(
            patch_size=4, window_size=7, embed_dim=96, depths=(2, 2, 6, 2), num_heads=(3, 6, 12, 24), **kwargs),

        'swin_base_patch4_window12_384_in22k' : dict(
            patch_size=4, window_size=12, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32), **kwargs),

        'swin_base_patch4_window7_224_in22k' : dict(
            patch_size=4, window_size=7, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32), **kwargs),

        'swin_large_patch4_window12_384_in22k' : dict(
            patch_size=4, window_size=12, embed_dim=192, depths=(2, 2, 18, 2), num_heads=(6, 12, 24, 48), **kwargs),

        'swin_large_patch4_window7_224_in22k' : dict(
            patch_size=4, window_size=7, embed_dim=192, depths=(2, 2, 18, 2), num_heads=(6, 12, 24, 48), **kwargs)
    }
    return model_architectures[name], config



from .hyperparameter import HyperParameter
from .hyperparameter import param_scope, reads, writes, all_params
from .hyperparameter import auto_param, set_auto_param_callback

__all__ = [
    'HyperParameter', \
    'param_scope', 'reads', 'writes', 'all_params', \
    'auto_param', 'set_auto_param_callback'
]



# coding=utf-8
# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import os
import torch
import torch.distributed as dist

from typing import Union, Any, Optional
from pathlib import Path
from collections.abc import Iterable
from torch import nn
from torch import optim
from torch.optim import Optimizer
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset
import torch.multiprocessing as mp

from towhee.data.dataset.dataset import TowheeDataSet, TorchDataSet

from towhee.trainer.callback import TensorBoardCallBack, ProgressBarCallBack, PrintCallBack, ModelCheckpointCallback, \
    EarlyStoppingCallback, TrainerControl, Callback
from towhee.trainer.metrics import get_metric_by_name
from towhee.trainer.modelcard import ModelCard, MODEL_CARD_NAME
from towhee.trainer.utils.trainer_utils import STATE_CHECKPOINT_NAME, MODEL_NAME, set_seed, reduce_value, \
    is_main_process, send_to_device, unwrap_model, _construct_loss_from_config, _construct_optimizer_from_config, \
    _construct_scheduler_from_config
from towhee.trainer.training_config import TrainingConfig
from towhee.utils.log import trainer_log
from towhee.trainer.optimization.optimization import get_scheduler, get_warmup_steps
from towhee.trainer.callback import CallbackList, _get_summary_writer_constructor

WEIGHTS_NAME = "pytorch_model.bin"
TEMP_INIT_WEIGHTS = "./temp_init_weights.pt"
CUSTOM = "custom_"
no_option_list = ["no", "null", "None", None, False]


def freeze_bn(model):
    classname = model.__class__.__name__
    if classname.find("BatchNorm") != -1:
        model.eval()


class Trainer:
    """
    A `Trainer` is used to train a pytorch model.

    Args:
        model (`nn.Module`):
            A pytorch model.
        training_config (`TrainingConfig`):
            A `TrainingConfig` instance can be loaded from a yaml file.
        train_dataset (`Union[Dataset, TowheeDataSet]`):
            It can be a kind of `torch.utils.data.dataset.Dataset` or `TowheeDataSet`
        eval_dataset (`Union[Dataset, TowheeDataSet]`):
            The same as `train_dataset`, and it is not strictly necessary if you do not want to eval.
        train_dataloader (`Union[DataLoader, Iterable]`):
            When the `train_dataloader` is passed in, trainer will use it to load data
            instead of constructing by input dataset.
        eval_dataloader (`Union[DataLoader, Iterable]`):
            The same as `train_dataloader`, and it is not strictly necessary also.
        model_card (`ModelCard`):
            Model card may contain other information of a model, and it is not strictly necessary.

    Examples:
        >>> import torch
        >>> import torchvision.models as models
        >>> from towhee.trainer.trainer import Trainer
        >>> from towhee.trainer.training_config import TrainingConfig
        >>> from towhee import dataset
        >>> from torchvision import transforms
        >>> import warnings
        >>> warnings.filterwarnings("ignore")
        >>> model = models.resnet18()
        >>> fake_transform = transforms.Compose([transforms.ToTensor()])
        >>> train_data = dataset('fake', size=2, transform=fake_transform)
        >>> val_data = dataset('fake', size=1, transform=fake_transform)
        >>> training_config = TrainingConfig(output_dir="train_res",
        ...                                  tensorboard=None,
        ...                                  epoch_num=2,
        ...                                  batch_size=1,
        ...                                  dataloader_num_workers=0,
        ...                                  print_steps=1)
        >>> trainer = Trainer(model, training_config, train_dataset=train_data, eval_dataset=val_data)
        >>> type(trainer)
        <class 'towhee.trainer.trainer.Trainer'>
        >>> trainer.train()  # Some values below are not necessarily reproducible
        2022-03-01 19:01:54,314 - 8601085440 - trainer.py-trainer:324 - WARNING: TrainingConfig(...)
        epoch=1/2, global_step=1, epoch_loss=7.107283592224121, epoch_metric=0.0
        epoch=1/2, global_step=2, epoch_loss=6.959554195404053, epoch_metric=0.0
        epoch=1/2, eval_global_step=0, eval_epoch_loss=6.694866180419922, eval_epoch_metric=0.0
        epoch=2/2, global_step=3, epoch_loss=6.165490627288818, epoch_metric=0.0
        epoch=2/2, global_step=4, epoch_loss=6.197325706481934, epoch_metric=0.0
        epoch=2/2, eval_global_step=1, eval_epoch_loss=6.0876030921936035, eval_epoch_metric=0.0
        >>> trainer.train(resume_checkpoint_path="train_res/epoch_1")
        2022-03-01 19:01:57,004 - 8601085440 - trainer.py-trainer:324 - WARNING: TrainingConfig(...)
        epoch=2/2, global_step=1, epoch_loss=6.165490627288818, epoch_metric=0.0
        epoch=2/2, global_step=2, epoch_loss=6.277336120605469, epoch_metric=0.0
        epoch=2/2, eval_global_step=0, eval_epoch_loss=6.097333908081055, eval_epoch_metric=0.0
        >>> trainer.save(path="another_save_path")
        >>> trainer.load(path="another_save_path")
        >>> trainer.epoch
        2
        >>> model = models.resnet18()
        >>> inputs = [torch.randn(1, 3, 224, 224), torch.Tensor([1]).type(torch.LongTensor)]
        >>> model.eval()  # turn on eval mode
        ResNet(
        ...
        )
        >>> trainer.evaluate_step(model, inputs)
        {'eval_step_loss': 7.10837459564209, 'eval_epoch_loss': 6.350093841552734, 'eval_epoch_metric': 0.0}
        >>> trainer.update_metrics(model, inputs, torch.Tensor(1))
        (5.2800750732421875, 0.0)
        >>> trainer.compute_metric(model, inputs)
        0.0
        >>> trainer.evaluate(model, {"epoch": 1, "eval_global_step": 1})
        epoch=1/2, eval_global_step=1, eval_epoch_loss=5.654547214508057, eval_epoch_metric=0.0
        {'epoch': 1, 'eval_global_step': 2, 'eval_step_loss': 7.526906967163086, 'eval_epoch_loss': 5.654547214508057, 'eval_epoch_metric': 0.0}
        >>> trainer.predict(inputs[0]).shape
        torch.Size([1, 1000])
        >>> model.train()  # turn on train mode
        ResNet(
        ...
        )
        >>> trainer.train_step(model, inputs)
        {'step_loss': 7.10837459564209, 'epoch_loss': 5.862236976623535, 'epoch_metric': 0.0}
        >>> trainer.compute_loss(model, inputs)
        tensor(7.1084, grad_fn=<NllLossBackward>)
        >>>
        >>> from towhee.trainer.callback import Callback
        >>> from typing import Dict
        >>> class MyCallback(Callback):
        ...     def on_eval_begin(self, logs: Dict) -> Dict:
        ...         print("on_eval_begin...")
        ...
        >>> my_callback = MyCallback()
        >>> trainer.add_callback(my_callback)
        >>> trainer.evaluate(model, logs={"epoch": 1, "eval_global_step": 1})
        on_eval_begin...
        epoch=1/2, eval_global_step=1, eval_epoch_loss=6.070321083068848, eval_epoch_metric=0.0
        {'epoch': 1, 'eval_global_step': 2, 'eval_step_loss': 7.526906967163086, 'eval_epoch_loss': 6.070321083068848, 'eval_epoch_metric': 0.0}
    """

    def __init__(
            self,
            model: nn.Module = None,
            training_config: TrainingConfig = None,
            train_dataset: Union[Dataset, TowheeDataSet] = None,
            eval_dataset: Union[Dataset, TowheeDataSet] = None,
            train_dataloader: Union[DataLoader, Iterable] = None,
            eval_dataloader: Union[DataLoader, Iterable] = None,
            model_card: ModelCard = None
    ):

        if training_config is None:
            output_dir = "tmp_trainer"
            trainer_log.warning("No `TrainingConfig` passed.")
            training_config = TrainingConfig(output_dir=output_dir)
        self.configs = training_config

        if model is None:
            raise RuntimeError("`Trainer` requires either a `model` or `model_init` argument")

        if isinstance(train_dataset, Dataset):
            self.train_dataset = train_dataset
        elif isinstance(train_dataset, TowheeDataSet):
            self.train_dataset = train_dataset.dataset

        self.eval_dataset = eval_dataset
        self.model = model
        self.model_card = model_card
        self.optimizer = None
        self.override_optimizer = False
        self.lr_scheduler_type = self.configs.lr_scheduler_type
        self.lr_scheduler = None
        self.lr_value = self.configs.lr
        self.metric = None
        self.loss = None
        self.override_loss = False
        self.loss_value = 0.0
        self.callbacks = CallbackList()
        self.loss_metric = None
        self.metric_value = 0.0
        self.epoch = 0
        self.train_dataloader = train_dataloader
        # self.train_sampler = None
        self.eval_dataloader = eval_dataloader
        self.distributed = False

        Path(self.configs.output_dir).mkdir(exist_ok=True)
        if not isinstance(self.model_card, ModelCard):
            self.model_card = ModelCard()

        if self.model_card.model_name is None:
            self.model_card.model_name = type(self.model).__name__
        self.model_card.model_architecture = str(self.model)
        self.model_card.training_config = self.configs

    def train(self, resume_checkpoint_path: Optional[str] = None):
        """
        Start to train.
        Args:
            resume_checkpoint_path (`int`):
                The path to start resume training.
        """
        if self.configs.device_str == "cuda":
            self.distributed = True
            self._spawn_train_process(resume_checkpoint_path)
        else:
            self.distributed = False
            self.run_train(None, None, resume_checkpoint_path)

    def _spawn_train_process(self, resume_checkpoint_path: Optional[str]):
        world_size = torch.cuda.device_count()
        mp.spawn(self.run_train,
                 args=(world_size, resume_checkpoint_path),
                 nprocs=world_size,  # opt.world_size,
                 join=True)

    def _init_distributed(self, rank: int, world_size: int):
        if self.distributed:
            if torch.cuda.is_available() is False:
                raise EnvironmentError("not find GPU device for training.")
            os.environ["MASTER_ADDR"] = "localhost"
            os.environ["MASTER_PORT"] = "12355"
            trainer_log.warning("_init_distributed(), rank=%s", rank)
            torch.cuda.set_device(rank)
            dist_backend = "nccl"
            dist_url = "env://"
            trainer_log.warning("| distributed init (rank %s): %s", rank, dist_url)
            dist.init_process_group(backend=dist_backend, init_method=dist_url,
                                    world_size=world_size, rank=rank)
            dist.barrier()

    def _load_before_train(self, resume_checkpoint_path: Optional[str], rank: Optional[int]):
        sync_bn = self.configs.sync_bn
        if resume_checkpoint_path is not None:
            # weights_dict = torch.load(weights_path, map_location=device)
            # load_weights_dict = {k: v for k, v in weights_dict.items()
            #                      if model.state_dict()[k].numel() == v.numel()}
            # model.load_state_dict(load_weights_dict, strict=False)
            self.load(resume_checkpoint_path)
        else:  # if using multi gpu and not resume, must keep model replicas in all processes are the same
            if self.distributed:
                # checkpoint_path = os.path.join(tempfile.gettempdir(), TEMP_INIT_WEIGHTS)
                checkpoint_path = TEMP_INIT_WEIGHTS
                if rank == 0:
                    torch.save(self.model.state_dict(), checkpoint_path)
                dist.barrier()
                self.model.load_state_dict(torch.load(checkpoint_path, map_location=self.configs.device))
        if self.distributed:
            self.model = torch.nn.parallel.DistributedDataParallel(self.model, device_ids=[rank])
            if sync_bn:
                self.model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self.model).to(self.configs.device)

    def _create_logs(self):
        logs = {"global_step": 0, "epoch": self.epoch}
        if self.configs.eval_strategy not in no_option_list:
            logs["eval_global_step"] = 0
        return logs

    def prepare_inputs(self, inputs: Any):
        return send_to_device(inputs, self.configs.device)

    def run_train(self, rank: int = None, world_size: int = None, resume_checkpoint_path: str = None):
        """
        Main training entry point.
        It is not recommended for users to use it unless over rewriting Trainer.
        Instead, it is recommended to use `trainer.train()` to start training.

        Args:
            rank (`int`):
                Process rank when using multi gpus.
            world_size (`int`):
                Total processes count.
            resume_checkpoint_path (`str`):
                Last checkpoint path.
        """
        set_seed(self.configs.seed)
        self._init_distributed(rank, world_size)
        self.model = self.model.to(self.configs.device)
        model = self.model
        self.trainercontrol = TrainerControl()
        self._load_before_train(resume_checkpoint_path, rank)
        # Keeping track whether we can can len() on the dataset or not
        # train_dataset_is_sized = isinstance(self.train_dataset, collections.abc.Sized)

        train_dataloader = self.get_train_dataloader()

        total_train_batch_size = self.configs.train_batch_size
        # if train_dataset_is_sized:
        # num_update_steps_per_epoch = len(train_dataloader)
        num_update_steps_per_epoch = len(self.train_dataset)
        num_update_steps_per_epoch = max(num_update_steps_per_epoch, 1)

        train_last_epoch = self.epoch
        num_train_epochs = math.ceil(self.configs.epoch_num - train_last_epoch)
        num_train_steps = math.ceil(num_train_epochs * num_update_steps_per_epoch)
        self.setup_before_train(num_training_steps=num_train_steps, init_lr=self.lr_value)
        trainer_log.info("***** Running training *****")
        trainer_log.info("  Num Epochs = %d", num_train_epochs)
        trainer_log.info("  Total train batch size  = %d", total_train_batch_size)
        trainer_log.info("****************************")

        if is_main_process():
            trainer_log.warning(self.configs)

        logs = self._create_logs()
        self.callbacks.on_train_begin(logs)
        for epoch in range(train_last_epoch + 1, self.configs.epoch_num + 1):
            self.epoch = logs["epoch"] = epoch
            self.set_train_mode(model)
            self._reset_controller()
            self.optimizer.zero_grad()
            # if self.distributed: # todo
                # self.train_sampler.set_epoch(self.epoch)
            # batch_loss_sum = 0.0
            self.callbacks.on_epoch_begin(self.epoch, logs)
            self.loss_metric.reset()
            self.metric.reset()
            for step, inputs in enumerate(train_dataloader):
                self.callbacks.on_train_batch_begin(inputs, logs)
                inputs = self.prepare_inputs(inputs)
                step_logs = self.train_step(model, inputs)  # , train_dataloader)
                logs["lr"] = self.lr_scheduler.get_last_lr()[0]
                logs["global_step"] += 1
                logs.update(step_logs)
                self.callbacks.on_train_batch_end(tuple(inputs), logs)
                self._may_evaluate(model, logs, step)
            self._may_evaluate(model, logs)
            self.callbacks.on_epoch_end(self.epoch, logs)
            self.loss_value = logs["epoch_loss"]
            self.metric_value = logs["epoch_metric"]
            self.lr_value = logs["lr"]
            self._create_training_summary(
                finetuned_from=resume_checkpoint_path,
                resumed_from_epoch=train_last_epoch if train_last_epoch != 0 else None,
                num_train_epochs=self.epoch - train_last_epoch,
                current_epoch=self.epoch,
                end_lr=self.lr_value,
                loss={"type": self.configs.loss, "value": round(self.loss_value, 3)},
                metric={"type": self.configs.metric, "value": round(self.metric_value, 3)}
            )
            if self.trainercontrol.should_training_stop:
                break
            if self.trainercontrol.should_save:
                self.save(
                    path=Path(self.configs.output_dir).joinpath("epoch_" + str(self.epoch)),
                    overwrite=self.configs.overwrite_output_dir
                )
        trainer_log.info("\nTraining completed.\n")
        self._cleanup_distributed(rank)
        self.callbacks.on_train_end(logs)
        self.save(
            path=Path(self.configs.output_dir).joinpath("final_epoch"),
            overwrite=self.configs.overwrite_output_dir
        )

    def set_train_mode(self, model: nn.Module):
        """
        Convert the model to training mode.

        Args:
            model (`nn.Module`):
        """
        model.train()
        if self.configs.freeze_bn:
            model.apply(freeze_bn)

    def _create_training_summary(self, **kwargs):
        training_summary = dict(kwargs)
        self.model_card.training_summary = training_summary

    def _may_evaluate(self, model: nn.Module, logs: dict, step: int = -1):
        if step != -1:  # step end
            if self.configs.eval_strategy in ["step", "steps"]:
                assert self.configs.eval_steps > 0, "self.configs.eval_steps must be a positive int number"
            if self.configs.eval_strategy in ["step", "steps"] and step % self.configs.eval_steps == 0:
                eval_logs = self.evaluate(model, logs)
                logs.update(eval_logs)
        else:  # epoch end
            if self.configs.eval_strategy in ["epoch", "eval_epoch"]:
                eval_logs = self.evaluate(model, logs)
                logs.update(eval_logs)

    @torch.no_grad()
    def evaluate_step(self, model: nn.Module, inputs: Any) -> dict:
        """
        One batch step when evaluating.

        Args:
            model (`nn.Module`):
                Pytorch model.
            inputs (`Any`):
                Input Tensor or any kind of collection made up by tensors.

        Returns:
            (`dict`)
                Evaluate logs dict.
        """
        inputs = self.prepare_inputs(inputs)
        step_loss = self.compute_loss(model, inputs)
        step_loss = reduce_value(step_loss, average=True)
        step_loss = step_loss.detach()

        loss_metric, epoch_metric = self.update_metrics(model, inputs, step_loss, training=False)

        step_logs = {"eval_step_loss": step_loss.item(), "eval_epoch_loss": loss_metric,
                     "eval_epoch_metric": epoch_metric}
        return step_logs

    @torch.no_grad()
    def update_metrics(self, model: nn.Module, inputs: Any, step_loss: torch.Tensor, training: bool = True) -> tuple:
        """
        Update the loss and metric in one epoch.
        When restart a new epoch, the epoch_loss and epoch metric will be clear.

        Args:
            model (`nn.Module`):
                Pytorch model.
            inputs (`Any`):
                Torch tensor or any kind of collection made up by tensors.
            step_loss (`torch.Tensor`):
                One batch step loss.
            training (`bool`):
                Whether it's training mode now.

        Returns:
            (`tuple`)
                Epoch loss and epoch metric.
        """
        self.loss_metric.update(send_to_device(step_loss, self.configs.device))
        loss_metric = self.loss_metric.compute().item()
        if self.configs.eval_strategy == "eval_epoch" and training:
            epoch_metric = 0
        else:
            epoch_metric = self.compute_metric(model, inputs)
        return loss_metric, epoch_metric

    @torch.no_grad()
    def compute_metric(self, model: nn.Module, inputs: Any) -> float:
        """
        Compute the step metric.
        It is recommended to subclass `Trainer` and override this method when deal with custom metric in custom task.
        When it is overridden, another method `compute_loss()` often needs to be overridden.

        Args:
            model (`nn.Module`):
                Pytorch model.
            inputs (`Any`):
                Input tensor collection.

        Returns:
            (`float`)
                Epoch average metric.
        """
        model.eval()
        epoch_metric = None
        labels = inputs[1]
        outputs = model(inputs[0])
        if self.metric is not None:
            self.metric.update(send_to_device(outputs, self.configs.device),
                               send_to_device(labels, self.configs.device))
            epoch_metric = self.metric.compute().item()
        return epoch_metric

    @torch.no_grad()
    def evaluate(self, model: nn.Module, logs: dict) -> dict:
        """
        Evaluate the model.

        Args:
            model (`nn.Module`):
                Pytorch model.
            logs (`dict`):
                Logs dict.

        Returns:
            (`dict`)
                The new logs dict with evaluate values.
        """
        model.eval()
        self.callbacks.on_eval_begin(logs)
        self.metric.reset()
        eval_dataloader = self.get_eval_dataloader()
        if eval_dataloader is None:
            trainer_log.warning("eval_dataloader is None!")
            return logs
        for _, inputs in enumerate(eval_dataloader):
            self.callbacks.on_eval_batch_begin(inputs, logs)
            inputs = send_to_device(inputs, self.configs.device)
            step_logs = self.evaluate_step(model, inputs)
            logs.update(step_logs)
            self.callbacks.on_eval_batch_end(tuple(inputs), logs)
            logs["eval_global_step"] += 1
        self.callbacks.on_eval_end(logs)
        return logs

    @torch.no_grad()
    def predict(self, inputs: Any) -> Any:
        """
        Do prediction. The eval mode model passes in the input value and get the outputs.

        Args:
            inputs (`Any`):
                Inference inputs by the model.

        Returns:
            (`Any`)
                Output result.
        """
        self.model.eval()
        return self.model(inputs)

    def train_step(self, model: nn.Module, inputs: Any) -> dict:
        """
        The training batch step.
        It contains computing step loss, loss backpropagation, doing step optimization, computing metric.

        Args:
            model (`nn.Module`):
                Pytorch model.
            inputs (`Any`):
                Pytorch tensor or tensor collection.

        Returns:
            (`dict`)
                Step logs which contains the step loss and metric infos.
        """
        step_loss = self.compute_loss(model, inputs)
        step_loss = reduce_value(step_loss, average=True)
        step_loss.backward()
        step_loss = step_loss.detach()

        loss_metric, epoch_metric = self.update_metrics(model, inputs, step_loss, training=True)

        self.optimizer.step()
        self.lr_scheduler.step()
        self.optimizer.zero_grad()
        step_logs = {"step_loss": step_loss.item(), "epoch_loss": loss_metric, "epoch_metric": epoch_metric}
        return step_logs

    def _cleanup_distributed(self, rank: int):
        if self.distributed:
            if rank == 0:
                if Path(TEMP_INIT_WEIGHTS).exists() is True:
                    os.remove(TEMP_INIT_WEIGHTS)
            # if is_main_process():
            # self.model = unwrap_model(self.model)
            dist.destroy_process_group()

    def compute_loss(self, model: nn.Module, inputs: Any):
        """
        Compute the step loss.
        It is recommended to subclass `Trainer` and override this method when deal with custom loss in custom task.
        When it is overridden, another method `compute_metric()` often needs to be overridden.

        Args:
            model (`nn.Module`):
                Pytorch model.
            inputs (`Any`):
                Model inputs when training.

        Returns:
            (`Any`)
                Loss values with `grad_fn`.
        """
        self.set_train_mode(model)
        labels = inputs[1]
        outputs = model(inputs[0])
        loss = self.loss(outputs, labels)
        return loss

    def push_model_to_hub(self):
        # todo
        pass

    def add_callback(self, callback: Callback, singleton: bool = True):
        """
        Users can add their custom callbacks into the trainer.

        Args:
            callback (`Callback`):
                Custom callback.
            singleton (`bool`):
                Whether this kind of callback is singleton.
                When singleton, the same class instance in `trainer.callbacks` will be replaced.
        """
        self.callbacks.add_callback(callback, singleton)

    def set_optimizer(self, optimizer: optim.Optimizer, optimizer_name: str = None):
        """
        Set custom optimizer

        Args:
            optimizer (`optim.Optimizer`):
                User's custom optimizer instance.
            optimizer_name (`str`):
                The optimizer string in training config, if it is `None`, the string will be a default value.

        Examples:
            >>> from towhee.trainer.trainer import Trainer
            >>> from typing import Optional, Callable
            >>> from torch import optim
            >>> import torchvision.models as models
            >>> model = models.resnet18()
            >>> trainer = Trainer(model)
            2022-03-01 17:22:52,306 - 8614221312 - trainer.py-trainer:173 - WARNING: No `TrainingConfig` passed.
            >>> class MyOptimizer(optim.Optimizer):
            ...     def step(self, closure: Optional[Callable[[], float]]=...) -> Optional[float]:
            ...         print('my step...')
            ...
            >>> my_optimizer = MyOptimizer(model.parameters(), defaults={})
            >>> trainer.set_optimizer(my_optimizer)
            >>> type(trainer.optimizer)
            <class '__main__.MyOptimizer'>
        """
        self.override_optimizer = True
        self.configs.optimizer = CUSTOM if optimizer_name is None else optimizer_name
        self.optimizer = optimizer

    def set_loss(self, loss: Any, loss_name: str = None):
        """
        Set custom loss

        Args:
            loss (`Any`):
                User's custom loss instance.
            loss_name (`str`):
                The loss string in training config, if it is `None`, the string will be a default value.

        Examples:
            >>> from towhee.trainer.trainer import Trainer
            >>> import torchvision.models as models
            >>> import torch
            >>> model = models.resnet18()
            >>> trainer = Trainer(model)
            2022-03-01 17:34:36,873 - 8605304320 - trainer.py-trainer:173 - WARNING: No `TrainingConfig` passed.
            >>> class MyTripletLossFunc(torch.nn.Module):
            ...     def forward(self):
            ...         print('forward...')
            ...         return 0
            ...
            >>> my_loss = MyTripletLossFunc()
            >>> trainer.set_loss(my_loss)
            >>> type(trainer.loss)
            <class '__main__.MyTripletLossFunc'>
        """
        self.override_loss = True
        self.configs.loss = CUSTOM if loss_name is None else loss_name
        self.loss = loss

    def _get_num_workers(self):
        if self.configs.dataloader_num_workers == -1:
            num_workers = min([os.cpu_count(), self.configs.batch_size if self.configs.batch_size > 1 else 0, 8])
        else:
            num_workers = self.configs.dataloader_num_workers
        if is_main_process():
            trainer_log.info("num_workers=%s", num_workers)
        return num_workers

    def get_train_dataloader(self) -> DataLoader:
        """
        Get the training dataloader.

        Returns:
            ('Optional[DataLoader]')
                The dataloader to fit data to train the model.
        """
        if self.train_dataloader is not None:
            return self.train_dataloader
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")
        if isinstance(self.train_dataset, TorchDataSet):
            self.train_dataset = self.train_dataset.dataset
        num_workers = self._get_num_workers()
        # if isinstance(self.train_dataset, IterableDataset):
        #     return DataLoader(
        #         self.train_dataset,
        #         batch_size=self.configs.train_batch_size,
        #     )
        if not self.distributed:
            return DataLoader(
                self.train_dataset,
                batch_size=self.configs.train_batch_size,
                shuffle=True,
                num_workers=num_workers,  # self.configs.dataloader_num_workers,
                pin_memory=self.configs.dataloader_pin_memory,
                drop_last=self.configs.dataloader_drop_last
            )
        else:
            # self.train_sampler = torch.utils.data.distributed.DistributedSampler(self.train_dataset)
            train_sampler = torch.utils.data.distributed.DistributedSampler(self.train_dataset)
            train_batch_sampler = torch.utils.data.BatchSampler(
                train_sampler, self.configs.batch_size, drop_last=True)
            return torch.utils.data.DataLoader(self.train_dataset,
                                               batch_sampler=train_batch_sampler,
                                               num_workers=num_workers,  # self.configs.dataloader_num_workers,
                                               pin_memory=self.configs.dataloader_pin_memory,
                                               )

    def get_eval_dataloader(self) -> Optional[DataLoader]:
        """
        Get the eval dataloader.

        Returns:
            (`Optional[DataLoader]`)
                The dataloader to fit data to eval the model.
        """
        if self.eval_dataloader is not None:
            return self.eval_dataloader
        if self.eval_dataset is None:
            trainer_log.warning("Trainer: eval requires a train_dataset.")
            return None
        if isinstance(self.eval_dataset, TorchDataSet):
            self.eval_dataset = self.eval_dataset.dataset
        # if isinstance(self.eval_dataset, IterableDataset):
        #     return DataLoader(
        #         self.eval_dataset,
        #         batch_size=self.configs.eval_batch_size,
        #     )
        num_workers = self._get_num_workers()
        if not self.distributed:
            return DataLoader(
                self.eval_dataset,
                batch_size=self.configs.eval_batch_size,
                num_workers=num_workers,  # self.configs.dataloader_num_workers,
                pin_memory=self.configs.dataloader_pin_memory,
                drop_last=self.configs.dataloader_drop_last
            )
        else:
            eval_sampler = torch.utils.data.distributed.DistributedSampler(self.eval_dataset)
            eval_batch_sampler = torch.utils.data.BatchSampler(
                eval_sampler, self.configs.batch_size, drop_last=True)
            return torch.utils.data.DataLoader(self.eval_dataset,
                                               batch_sampler=eval_batch_sampler,
                                               num_workers=num_workers,  # self.configs.dataloader_num_workers,
                                               pin_memory=self.configs.dataloader_pin_memory,
                                               )

    def setup_before_train(self, num_training_steps: int, init_lr: float):
        """
        Setup some configs before training.

        Args:
            num_training_steps (`int`):
                All training steps in all training loops.
            init_lr (`float`):
                Start learning rate.
        """
        self._create_optimizer(init_lr=init_lr)
        self._create_loss()
        self._create_metric()
        self._create_scheduler(num_training_steps=num_training_steps, optimizer=self.optimizer)
        self._create_callbacks()

    def _create_callbacks(self):
        # print or progressbar
        if self.configs.print_steps is None:
            self.callbacks.add_callback(ProgressBarCallBack(total_epoch_num=self.configs.epoch_num,
                                                            train_dataloader=self.get_train_dataloader()))
        else:
            self.callbacks.add_callback(PrintCallBack(total_epoch_num=self.configs.epoch_num,
                                                      step_frequency=self.configs.print_steps))
        # early stop
        if self.configs.early_stopping not in no_option_list:
            self.callbacks.add_callback(EarlyStoppingCallback(self.trainercontrol, **self.configs.early_stopping))
        # save checkpoint
        if self.configs.model_checkpoint not in no_option_list:
            self.callbacks.add_callback(ModelCheckpointCallback(self.trainercontrol, **self.configs.model_checkpoint))
        # tensorboard
        if self.configs.tensorboard not in no_option_list:
            summary_writer_constructor = _get_summary_writer_constructor()
            if summary_writer_constructor is not None:
                self.callbacks.add_callback(
                    TensorBoardCallBack(summary_writer_constructor,
                                        **self.configs.tensorboard))

    def _create_metric(self):
        self.metric = get_metric_by_name(self.configs.metric)
        self.metric.to(self.configs.device)
        self.loss_metric = get_metric_by_name("MeanMetric")
        self.loss_metric.to(self.configs.device)

    def _create_loss(self):
        if self.override_loss is True:
            return
        self.loss = _construct_loss_from_config(torch.nn.modules.loss, self.configs.loss)

    def _create_optimizer(self, init_lr: float):
        if self.override_optimizer is True:
            return
        self.optimizer = _construct_optimizer_from_config(
            optim,
            self.configs.optimizer,
            model=self.model,
        )
        for param in self.optimizer.param_groups:
            param.setdefault("initial_lr", init_lr)
        self.optimizer.lr = init_lr

    def _create_scheduler(self, num_training_steps: int, optimizer: torch.optim.Optimizer = None):
        if isinstance(self.configs.lr_scheduler_type, str):
            self.lr_scheduler = get_scheduler(
                self.configs.lr_scheduler_type,
                optimizer=self.optimizer if optimizer is None else optimizer,
                num_warmup_steps=get_warmup_steps(num_training_steps, self.configs.warmup_steps,
                                                  self.configs.warmup_ratio),
                num_training_steps=num_training_steps,
            )
        elif isinstance(self.configs.lr_scheduler_type, dict):
            self.lr_scheduler = _construct_scheduler_from_config(optimizer,
                                                                 torch.optim.lr_scheduler,
                                                                 self.configs.lr_scheduler_type)
        return self.lr_scheduler

    def load(self, path: str):
        """
        Load a model from the path.

        Args:
            path (`str`):
                The folder path containing the model's checkpoints.
        """
        state_path = Path(path).joinpath(STATE_CHECKPOINT_NAME)
        model_path = Path(path).joinpath(MODEL_NAME)
        # modelcard_path = Path(path).joinpath(MODEL_CARD_NAME)
        trainer_log.info("Loading from previous checkpoint: %s", model_path)
        state_checkpoint = torch.load(state_path, map_location=self.configs.device)
        model_checkpoint = torch.load(model_path, map_location=self.configs.device)
        self.model.load_state_dict(model_checkpoint)
        if isinstance(self.optimizer, Optimizer) and state_checkpoint["optimizer_state_dict"]:
            self.optimizer.load_state_dict(state_checkpoint["optimizer_state_dict"])
        if self.lr_scheduler and state_checkpoint["lr_scheduler_state_dict"]:
            self.lr_scheduler.load_state_dict(state_checkpoint["lr_scheduler_state_dict"])
        if "end_lr" not in state_checkpoint:
            return 0
        self.lr_value = state_checkpoint["end_lr"]
        if "epoch" not in state_checkpoint:
            return 0
        self.epoch = state_checkpoint["epoch"]
        self.loss_value = state_checkpoint["loss_value"]
        self.metric_value = state_checkpoint["metric_value"]

    def save(self, path, overwrite=True):
        """
        Save the checkpoint information in a folder.

        Args:
            path (`Any`):
                The folder path containing the model's checkpoints.
            overwrite (`bool`):
                If True, it will overwrite the same name path when existing.

        Raises:
            (`FileExistsError`)
                If `overwrite` is False, when there already exists a path, it will raise Error.
        """
        if is_main_process():
            if not overwrite:
                if Path(path).exists():
                    raise FileExistsError("File already exists: ", str(Path(path).resolve()))
            Path(path).mkdir(exist_ok=True)
            state_path = Path(path).joinpath(STATE_CHECKPOINT_NAME)
            model_path = Path(path).joinpath(MODEL_NAME)
            modelcard_path = Path(path).joinpath(MODEL_CARD_NAME)
            trainer_log.info("save model_path: %s", model_path)
            optimizer_state_dict = None
            lr_scheduler_state_dict = None
            if isinstance(self.optimizer, Optimizer):  # if created
                optimizer_state_dict = self.optimizer.state_dict()
            if self.lr_scheduler is not None:
                lr_scheduler_state_dict = self.lr_scheduler.state_dict()
            torch.save({
                "epoch": self.epoch,
                "optimizer_state_dict": optimizer_state_dict,
                "lr_scheduler_state_dict": lr_scheduler_state_dict,
                "loss_value": self.loss_value,
                "metric_value": self.metric_value,
                "end_lr": self.lr_value
            }, state_path)
            torch.save(unwrap_model(self.model).state_dict(), model_path)
            if isinstance(self.model_card, ModelCard):
                self.model_card.save_model_card(modelcard_path)
            else:
                trainer_log.warning("model card is None.")

    def _reset_controller(self):
        self.trainercontrol.should_save = False
        self.trainercontrol.should_training_stop = False
        self.trainercontrol.should_log = False
        self.trainercontrol.should_evaluate = False
        self.trainercontrol.should_epoch_stop = False



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
from typing import List

import torch
from torch import nn

from towhee.models.utils.create_conv2d_pad import create_conv2d_pad

def _split_channels(num_chan: int, num_groups: int) -> List[int]:
    split = [num_chan // num_groups for _ in range(num_groups)]
    split[0] += num_chan - sum(split)
    return split

class MixedConv2d(nn.ModuleDict):
    """
    Mixed Grouped Convolution

    Based on MDConv and GroupedConv in MixNet impl:
    https://github.com/tensorflow/tpu/blob/master/models/official/mnasnet/mixnet/custom_layers.py

    Args:
        in_channels (`int`):
            Number of channels in the input image
        out_channels (`int`):
            Number of channels produced by the convolution.
        kernel_size (`int`):
            Size of the convolving kernel.
        stride (`int`):
            Stride of the convolution.
        padding (`str`):
            Padding added to all four sides of the input.
        dilation (`int`):
            Spacing between kernel elements.
        depthwise (`bool`):
            If True, use depthwise convolution.
    """
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int=3,
                 stride: int=1, padding: str='', dilation: int=1, depthwise: bool=False, **kwargs) -> None:
        super().__init__()

        kernel_size = kernel_size if isinstance(kernel_size, list) else [kernel_size]
        num_groups = len(kernel_size)
        in_splits = _split_channels(in_channels, num_groups)
        out_splits = _split_channels(out_channels, num_groups)
        self.in_channels = sum(in_splits)
        self.out_channels = sum(out_splits)
        for idx, (k, in_ch, out_ch) in enumerate(zip(kernel_size, in_splits, out_splits)):
            conv_groups = in_ch if depthwise else 1
            # use add_module to keep key space clean
            self.add_module(
                str(idx),
                create_conv2d_pad(
                    in_ch, out_ch, k, stride=stride,
                    padding=padding, dilation=dilation, groups=conv_groups, **kwargs)
            )
        self.splits = in_splits

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_split = torch.split(x, self.splits, 1)
        x_out = [c(x_split[i]) for i, c in enumerate(self.values())]
        x = torch.cat(x_out, 1)
        return x



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
from pathlib import Path, PosixPath
from typing import Union
import numpy as np
from io import BytesIO
from zipfile import ZipFile
from urllib.request import urlopen

from towhee.types import Image
from towhee.utils.log import engine_log
from towhee.utils.repo_normalize import RepoNormalize

try:
    import cv2
except ModuleNotFoundError as e:
    engine_log.error('cv2 not found, you can install via `pip install opencv-python`.')
    raise ModuleNotFoundError('cv2 not found, you can install via `pip install opencv-python`.') from e


def from_src(src: Union[str, PosixPath]) -> Image:
    """
    Load the image from url/path as towhee's Image object.

    Args:
        src (`Union[str, path]`):
            The path leads to the image.

    Returns:
        (`towhee.types.Image`)
            The image wrapepd as towhee's Image.
    """
    ndarray_img = cv2.imread(str(src))
    rgb_img = cv2.cvtColor(ndarray_img, cv2.COLOR_BGR2RGB)
    img = from_ndarray(rgb_img, 'RGB')

    return img


def from_zip(zip_src: Union[str, Path], pattern: str = '*.JPEG') -> Image:
    """
    Load the image.zip from url/path as towhee's Image object.

    Args:
        zip_src (`Union[str, path]`):
            The path leads to the image.
        pattern (`str`):
            The image pattern to extract.

    Returns:
        (`towhee.types.Image`)
            The image wrapepd as towhee's Image.
    """
    if RepoNormalize(str(zip_src)).url_valid():
        with urlopen(zip_src) as zip_file:
            zip_path = BytesIO(zip_file.read())
    else:
        zip_path = str(Path(zip_src).resolve())
    with ZipFile(zip_path, 'r') as zfile:
        file_list = zfile.namelist()
        path_list = glob.fnmatch.filter(file_list, pattern)
        img_list = []
        for path in path_list:
            data = zfile.read(path)
            ndarray_img = cv2.imdecode(np.frombuffer(data, np.uint8), 1)
            rgb_img = cv2.cvtColor(ndarray_img, cv2.COLOR_BGR2RGB)
            img = from_ndarray(rgb_img, 'RGB')
            img_list.append(img)

    return img_list


def from_ndarray(ndarray_img: np.ndarray, mode: str) -> Image:
    """
    Convert an image loaded by cv2 as ndarray into towhee.types.Image.

    Args:
        ndarray_img (`np.ndarray`):
            A image loaed by cv2 as ndarray.

    Returns:
        (`towhee.types.Image`)
            The image wrapepd as towhee Image.
    """
    img_bytes = ndarray_img.tobytes()
    img_width = ndarray_img.shape[1]
    img_height = ndarray_img.shape[0]
    img_channel = len(cv2.split(ndarray_img))
    img_mode = mode
    img_array = ndarray_img

    towhee_img = Image(img_bytes, img_width, img_height, img_channel, img_mode, img_array)

    return towhee_img


def to_ndarray(towhee_img: Image) -> np.ndarray:
    """
    Convert a towhee.types.Image into ndarray.

    The mode is same as `towhee_img`, use `towhee_img.mode` to get the information.

    Args:
        towhee_img (`towhee.types.Image`):
            A towhee image.

    Returns:
        (`np.ndarray`)
            The ndarray of the image, the mode is same as the `towhee_img`.
    """
    shape = (towhee_img.height, towhee_img.width, towhee_img.channel)
    data = towhee_img.image

    ndarray_img = np.ndarray(shape, np.uint8, data)

    return ndarray_img


def rgb2bgr(img: Union[np.ndarray, Image]) -> np.ndarray:
    """
    Convert an RGB image into ndarray of mode BGR.

    Args:
        img (`Union[np.ndarray, 'towhee.types.Image']`):
            An RGB image either in the form of ndarray or towhee's Image.

    Returns:
        (`np.ndarray`)
            An ndarray in the mode of BGR.
    """
    if isinstance(img, Image):
        if not img.mode.upper() == 'RGB':
            raise ValueError('The input image should be RGB mode.')
        else:
            rgb_img = img.array if isinstance(img.array, np.ndarray) else to_ndarray(img)

    elif isinstance(img, np.ndarray):
        rgb_img = img

    else:
        raise TypeError('Input image should either be an `np.ndarray` or a `towhee.types.Image`.')

    bgr_img = cv2.cvtColor(rgb_img, cv2.COLOR_RGB2BGR)

    return bgr_img



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.engine import register

# pylint: disable=import-outside-toplevel
# pylint: disable=invalid-name


@register(name='builtin/tensor_hstack')
def tensor_hstack(*arg):
    from scipy import sparse
    return sparse.hstack(arg)


@register(name='builtin/tensor_stack')
class tensor_stack:
    """
    Stack the sequence of inputs along a new axis.

    Args:
        axis (int, optional): the axis of the result array along which the inputs are
            stacked. Defaults to 0.

    Returns:
        ndarray: the stacked array.

    Examples:

    >>> import numpy as np
    >>> from towhee.functional import DataCollection
    >>> dc = (
    ...     DataCollection.range(10)
    ...         .map(lambda x: np.array([x]))
    ...         .batch(2)
    ...         .tensor_stack(axis=1)
    ... )
    >>> dc.to_list()
    [array([[0, 1]]), array([[2, 3]]), array([[4, 5]]), array([[6, 7]]), array([[8, 9]])]
    """

    def __init__(self, axis=0):
        self._axis = axis

    def __call__(self, x):
        import numpy as np

        return np.stack(x, axis=self._axis)


@register(name='builtin/tensor_unstack')
class tensor_unstack:
    """
    Unstack an array along given axis.

    Args:
        axis (int, optional): the axis along which to unstack the array. Defaults to 0.

    Returns:
        [ndarray]: sequence of result arrays.

    Examples:

    >>> import numpy as np
    >>> from towhee.functional import DataCollection
    >>> dc = (
    ...     DataCollection.range(3)
    ...         .map(lambda x: np.array([x, x*2]))
    ...         .tensor_unstack(axis=0)
    ... )
    >>> dc.to_list()
    [[array(0), array(0)], [array(1), array(2)], [array(2), array(4)]]
    """

    def __init__(self, axis=0):
        self._axis = axis

    def __call__(self, x):
        import numpy as np

        nsplit = x.shape[self._axis]
        shape = x.shape
        shape = [shape[i] for i in range(len(shape)) if i != self._axis]
        return list(
            map(lambda arr: arr.reshape(shape),
                np.split(x, nsplit, axis=self._axis)))


@register(name='builtin/tensor_concat')
class tensor_concat:
    """
    Concat the sequence of inputs along a axis (no new axis)

    Args:
        axis (int, optional): the axis alone which to concat inputs. Defaults to 0.

    Returns:
        ndarray: the stacked array.

    Examples:

    >>> import numpy as np
    >>> from towhee.functional import DataCollection
    >>> dc = (
    ...     DataCollection.range(10)
    ...         .map(lambda x: np.array([x]))
    ...         .batch(2)
    ...         .tensor_concat(axis=0)
    ... )
    >>> dc.to_list()
    [array([0, 1]), array([2, 3]), array([4, 5]), array([6, 7]), array([8, 9])]
    """

    def __init__(self, axis=0):
        self._axis = axis

    def __call__(self, x):
        import numpy as np
        return np.concatenate(list(x), axis=self._axis)


@register(name='builtin/tensor_split')
class tensor_split:
    """
    Split the input array along a axis.

    Args:
        axis (int, optional): the axis along which to split the array. Defaults to 0.

    Returns:
        [ndarray]: list of result arrays.

    Examples:

    >>> import numpy as np
    >>> from towhee.functional import DataCollection
    >>> dc = (
    ...     DataCollection.range(3)
    ...         .map(lambda x: np.array([[x, x*2]]))
    ...         .tensor_split(axis=1)
    ... )
    >>> dc.to_list()
    [[array([[0]]), array([[0]])], [array([[1]]), array([[2]])], [array([[2]]), array([[4]])]]
    """

    def __init__(self, axis=0):
        self._axis = axis

    def __call__(self, x):
        import numpy as np

        nsplit = x.shape[self._axis]
        return np.split(x, nsplit, axis=self._axis)


@register(name='builtin/tensor_normalize')
class tensor_normalize:
    """
    Normalize input tensor.

    Examples:

    >>> import numpy
    >>> from towhee.functional import DataCollection
    >>> dc = DataCollection([numpy.array([3, 4]), numpy.array([6,8])])
    >>> dc.tensor_normalize().to_list()
    [array([0.6, 0.8]), array([0.6, 0.8])]
    """

    def __init__(self, axis=0):
        self._axis = axis

    def __call__(self, x):
        import numpy as np

        return x / np.linalg.norm(x, axis=self._axis)


@register(name='builtin/tensor_reshape')
class tensor_reshape:

    def __init__(self, shape):
        self._shape = shape

    def __call__(self, x):

        return x.reshape(self._shape)


@register(name='builtin/tensor_random')
class tensor_random:
    """
    Return a random tensor filled with random numbers from [0.0, 1.0).

    Args:
        shape ([int], optional): tensor shape. Defaults to None.

    Returns:
        ndarray: output tensor.

    Examples:

    >>> from towhee.functional import DataCollection
    >>> DataCollection.range(5).tensor_random(shape=[1,2,3]).map(lambda x: x.shape).to_list()
    [(1, 2, 3), (1, 2, 3), (1, 2, 3), (1, 2, 3), (1, 2, 3)]
    """

    def __init__(self, shape):
        self._shape = shape

    def __call__(self, _):
        import numpy as np

        return np.random.random(self._shape)


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=True)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading


class AtomicCount:
    '''
    Atomic count, support thread-safe += operator.
    '''

    def __init__(self, count: int):
        self._count = count
        self._lock = threading.Lock()

    @property
    def count(self):
        return self._count

    def __iadd__(self, num: int) -> 'AtomicCount':
        with self._lock:
            self._count += num
        return self



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from typing import Any, Dict, List

from towhee.hparam import param_scope
from towhee.engine.uri import URI


def _get_default_namespace():
    with param_scope() as hp:
        return hp().towhee.default_namespace('anon')


class OperatorRegistry:
    """Operator Registry
    """

    REGISTRY: Dict[str, Any] = {}

    def __init__(self) -> None:
        pass

    @staticmethod
    def resolve(name: str) -> Any:
        """
        Resolve operator by name
        """
        for n in [
                name,
                '{}/{}'.format(_get_default_namespace(), name),
                '{}/{}'.format('builtin', name),
        ]:
            if n in OperatorRegistry.REGISTRY:
                return OperatorRegistry.REGISTRY[n]
        return None

    @staticmethod
    def register(
            name: str = None,
            input_schema=None,  # TODO: parse input_schema from code @jie.hou
            output_schema=None,
            flag=None):
        """
        Register a class, function, or callable as a towhee operator.

        Examples:

        1. register a function as operator

        >>> from towhee import register
        >>> @register
        ... def foo(x, y):
        ...     return x+y

        2. register a class as operator

        >>> @register
        ... class foo_cls():
        ...     def __init__(self, x):
        ...         self.x = x
        ...     def __call__(self, y):
        ...         return self.x + y

        By default, function/class name is used as operator name,
        which is used by the operator factory `towhee.ops` to invoke the operator.

        >>> from towhee import ops
        >>> op = ops.foo()
        >>> op(1, 2)
        3

        >>> op = ops.foo_cls(x=2)
        >>> op(3)
        5

        3. register operator with an alternative name:

        >>> @register(name='my_foo')
        ... def foo(x, y):
        ...     return x+y
        >>> ops.my_foo()(1,2)
        3

        Operator URI and Namespace: The URI (unique reference identifier) of an operator has two parts: namespace and name.
        The namespace helps identify one operator and group the operators into various kinds.
        We can specific the namespace when create an operator:

        >>> ops.anon.my_foo()(1,2)
        3

        `anon` is the default namespace to which an operator is registered if no namespace is specified.
        And it's also the default searching namespace for the operator factory.

        You can also specific the fullname, including namespace when register an operator:

        >>> @register(name='my_namespace/my_foo')
        ... def foo(x, y):
        ...     return x+y
        >>> ops.my_namespace.my_foo()(1,2)
        3

        Output Schema:

        >>> @register(name='my_foo', output_schema='value')
        ... def foo(x, y):
        ...     return x+y
        >>> from towhee.hparam import param_scope
        >>> with param_scope('towhee.need_schema=1'):
        ...     ops.my_foo()(1,2)
        Output(value=3)

        Flag: Each operator type, for example: NNOperator and PyOperator, has their own default `flag`:

        >>> from towhee.operator.base import Operator, NNOperator, PyOperator
        >>> from towhee.operator.base import OperatorFlag
        >>> @register
        ... class foo(NNOperator):
        ...     pass
        >>> foo().flag
        <OperatorFlag.REUSEABLE|STATELESS: 6>

        The default flag can be override by `register(flag=someflag)`:

        >>> @register(flag=OperatorFlag.EMPTYFLAG)
        ... class foo(NNOperator):
        ...     pass
        >>> foo().flag
        <OperatorFlag.EMPTYFLAG: 1>

        Args:
            name (str, optional): operator name, will use the class/function name if None.
            input_schema(NamedTuple, optional): input schema for the operator. Defaults to None.
            output_schema(NamedTuple, optional): output schema, will convert the operator output to NamedTuple if not None.
            flag ([OperatorFlag], optional): operator flag. Defaults to OperatorFlag.EMPTYFLAG.

        Returns:
            [type]: [description]
        """
        if callable(name):
            # the decorator is called directly without any arguments,
            # relaunch the register
            cls = name
            return OperatorRegistry.register()(cls)

        if output_schema is None:  # none output schema
            output_schema = namedtuple('Output', 'col0')
        if isinstance(output_schema, str):  # string schema 'col0 col1'
            output_schema = output_schema.split()
        if isinstance(output_schema, List):  # list schema ['col0', 'col1']
            output_schema = namedtuple('Output', output_schema)

        def wrapper(cls):
            metainfo = dict(input_schema=input_schema,
                            output_schema=output_schema,
                            flag=flag)

            nonlocal name
            name = URI(cls.__name__ if name is None else name).resolve_repo(
                _get_default_namespace())

            # wrap a callable to a class
            if not isinstance(cls, type) and callable(cls):
                func = cls
                cls = type(
                    cls.__name__, (object, ), {
                        '__call__': lambda _, *arg, **kws: func(*arg, **kws),
                        '__doc__': func.__doc__,
                    })

            if output_schema is not None:
                old_call = cls.__call__

                def wrapper_call(self, *args, **kws):
                    with param_scope() as hp:
                        need_schema = hp().towhee.need_schema(False)
                    if need_schema:
                        return output_schema(old_call(self, *args, **kws))
                    else:
                        return old_call(self, *args, **kws)

                cls.__call__ = wrapper_call
                cls.__abstractmethods__ = set()
            cls.metainfo = metainfo
            if flag is not None:
                cls.flag = property(lambda _: flag)
            OperatorRegistry.REGISTRY[name] = cls

            return cls

        return wrapper


if __name__ == '__main__':
    import doctest

    doctest.testmod(verbose=False)



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn
from typing import List, Tuple

from towhee.models.layers.multi_scale_attention import MultiScaleAttention
from towhee.models.layers.droppath import DropPath
from towhee.models.layers.mlp import Mlp
from towhee.models.layers.pool_attention import AttentionPool


class MultiScaleBlock(nn.Module):
    """
    A multiscale vision transformer block.
    Each block contains a multiscale attention layer and a Mlp layer.
    ::
                                      Input
                                        |-------------------+
                                        ↓                   |
                                       Norm                 |
                                        ↓                   |
                                MultiScaleAttention        Pool
                                        ↓                   |
                                     DropPath               |
                                        ↓                   |
                                    Summation ←-------------+
                                        |
                                        |-------------------+
                                        ↓                   |
                                       Norm                 |
                                        ↓                   |
                                       Mlp                 Proj
                                        ↓                   |
                                     DropPath               |
                                        ↓                   |
                                    Summation  ←------------+
    Args:
        dim(int):
            Input feature dimension.
        dim_out(int):
            Output feature dimension.
        num_heads(int):
            Number of heads in the attention layer.
        mlp_ratio(float):
            MLP ratio which controls the feature dimension in the hidden layer of the MLP block.
        qkv_bias(bool):
            If set to False, the qkv layer will not learn an additive bias.
        dropout_rate(float):
            DropOut rate. If set to 0, DropOut is disabled.
        droppath_rate(float):
            DropPath rate. If set to 0, DropPath is disabled.
        activation(nn.Module):
            Activation layer used in the MLP layer.
        norm_layer(nn.Module):
            Normalization layer.
        kernel_q(_size_3_t):
            Pooling kernel size for q. If pooling kernel size is 1 for all the dimensions.
        kernel_kv(_size_3_t):
            Pooling kernel size for kv. If pooling kernel size is 1 for all the dimensions, pooling is not used.
        stride_q(_size_3_t):
            Pooling kernel stride for q.
        stride_kv(_size_3_t):
            Pooling kernel stride for kv.
        pool_mode(nn.Module):
            Pooling mode.
        has_cls_embed(bool):
            If set to True, the first token of the input tensor should be a cls token.
            Otherwise, the input tensor does not contain a cls token. Pooling is not applied to the cls token.
        pool_first(bool):
            If set to True, pool is applied before qkv projection. Otherwise, pool is applied after qkv projection.
    """

    def __init__(
        self,
        dim,
        dim_out,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        dropout_rate=0.0,
        droppath_rate=0.0,
        activation=nn.GELU,
        norm_layer=nn.LayerNorm,
        kernel_q=(1, 1, 1),
        kernel_kv=(1, 1, 1),
        stride_q=(1, 1, 1),
        stride_kv=(1, 1, 1),
        pool_mode=nn.Conv3d,
        has_cls_embed=True,
        pool_first=False,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.dim_out = dim_out
        self.norm1 = norm_layer(dim)
        kernel_skip = [s + 1 if s > 1 else s for s in stride_q]
        stride_skip = stride_q
        padding_skip = [int(skip // 2) for skip in kernel_skip]
        self.attn = MultiScaleAttention(
            dim=dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            dropout_rate=dropout_rate,
            kernel_q=kernel_q,
            kernel_kv=kernel_kv,
            stride_q=stride_q,
            stride_kv=stride_kv,
            norm_layer=nn.LayerNorm,
            has_cls_embed=has_cls_embed,
            pool_mode=pool_mode,
            pool_first=pool_first,
        )
        self.drop_path = (
            DropPath(drop_prob=droppath_rate) if droppath_rate > 0.0 else nn.Identity()
        )
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.has_cls_embed = has_cls_embed
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            out_features=dim_out,
            act_layer=activation,
            drop=dropout_rate,
        )
        if dim != dim_out:
            self.proj = nn.Linear(dim, dim_out)

        self.pool_skip = (
            nn.MaxPool3d(kernel_skip, stride_skip, padding_skip, ceil_mode=False)
            if len(kernel_skip) > 0
            else None
        )

    def forward(
        self, x: torch.Tensor, thw_shape: List[int]
    ) -> Tuple[torch.Tensor, List[int]]:
        """
        Args:
            x(torch.Tensor):
                Input tensor.
            thw_shape(List):
                The shape of the input tensor (before flattening).
        """

        x_block, thw_shape_new = self.attn(self.norm1(x), thw_shape)
        atn = AttentionPool(
            pool=self.pool_skip,
            thw_shape=thw_shape,
            has_cls_embed=self.has_cls_embed
        )
        x_res, _ = atn(x)
        x = x_res + self.drop_path(x_block)
        x_norm = self.norm2(x)
        x_mlp = self.mlp(x_norm)
        if self.dim != self.dim_out:
            x = self.proj(x_norm)
        x = x + self.drop_path(x_mlp)
        return x, thw_shape_new



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
from torch import nn

class ConvMlp(nn.Module):
    """ MLP using 1x1 convs that keeps spatial dims
    """
    def __init__(
            self, in_features, hidden_features=None, out_features=None, act_layer=nn.ReLU, norm_layer=None, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=1, bias=True)
        self.norm = norm_layer(hidden_features) if norm_layer else nn.Identity()
        self.act = act_layer()
        self.fc2 = nn.Conv2d(hidden_features, out_features, kernel_size=1, bias=True)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.norm(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        return x



# Copyright 2021 biubug6 . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

#adapted from https://github.com/biubug6/Pytorch_Retinaface

import torch
from torch import nn

class ClassHead(nn.Module):
    """
    ClassHead

    RetinaFace head for classification branch.

    Args:
        inchannels (`int`):
            number of input channels.
        num_anchors (`int`):
            number of anchors.
    """
    def __init__(self,inchannels: int=512,num_anchors: int=3):
        super().__init__()
        self.num_anchors = num_anchors
        self.conv1x1 = nn.Conv2d(inchannels,self.num_anchors*2,kernel_size=(1,1),stride=1,padding=0)

    def forward(self,x: torch.FloatTensor):
        out = self.conv1x1(x)
        out = out.permute(0,2,3,1).contiguous()
        return out.view(out.shape[0], -1, 2)

class BboxHead(nn.Module):
    """
    BboxHead

    RetinaFace head for bounding box branch.

    Args:
        inchannels (`int`):
            number of input channels.
        num_anchors (`int`):
            number of anchors.
    """
    def __init__(self,inchannels: int=512,num_anchors: int=3):
        super().__init__()
        self.conv1x1 = nn.Conv2d(inchannels,num_anchors*4,kernel_size=(1,1),stride=1,padding=0)

    def forward(self,x: torch.FloatTensor):
        out = self.conv1x1(x)
        out = out.permute(0,2,3,1).contiguous()

        return out.view(out.shape[0], -1, 4)

class LandmarkHead(nn.Module):
    """
    LandmarkHead

    RetinaFace head for landmark branch.

        inchannels (`int`):
            number of input channels.
        num_anchors (`int`):
            number of anchors.
    """
    def __init__(self,inchannels: int=512,num_anchors: int=3):
        super().__init__()
        self.conv1x1 = nn.Conv2d(inchannels,num_anchors*10,kernel_size=(1,1),stride=1,padding=0)

    def forward(self,x: torch.FloatTensor):
        out = self.conv1x1(x)
        out = out.permute(0,2,3,1).contiguous()
        return out.view(out.shape[0], -1, 10)



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn


class ResNetBasic3DModule(nn.Module):
    """
    ResNet basic 3D stem module. It performs spatiotemporal Convolution, BN, and activation
    following by a spatiotemporal pooling.
                                        Conv3d
                                           ↓
                                     Normalization
                                           ↓
                                       Activation
                                           ↓
                                        Pool3d
    Args:
        conv (`nn.Module`):
            Convolutional module.
        norm (`nn.Module`):
            Normalization module.
        activation (`nn.Module`):
            Activation module.
        pool (`nn.Module`):
            Pooling module.
    """

    def __init__(
        self,
        *,
        conv=None,
        norm=None,
        activation=None,
        pool=None,
    ) -> None:
        super().__init__()
        self.conv = conv
        self.norm = norm
        self.activation = activation
        self.pool = pool

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        if self.norm is not None:
            x = self.norm(x)
        if self.activation is not None:
            x = self.activation(x)
        if self.pool is not None:
            x = self.pool(x)
        return x
